Package: startup
Version: 0.23.0
Title: Friendly R Startup Configuration
Depends: R (>= 2.14.0)
Suggests: commonmark, tools, tcltk
VignetteBuilder: startup
Authors@R: c(person("Henrik", "Bengtsson", role=c("aut", "cre", "cph"),
                                          email = "henrikb@braju.com"))
Description: Adds support for R startup configuration via '.Renviron.d' and '.Rprofile.d' directories in addition to '.Renviron' and '.Rprofile' files.  This makes it possible to keep private / secret environment variables separate from other environment variables.  It also makes it easier to share specific startup settings by simply copying a file to a directory.
License: LGPL (>= 2.1)
LazyLoad: TRUE
URL: https://henrikbengtsson.github.io/startup/,
        https://github.com/HenrikBengtsson/startup
BugReports: https://github.com/HenrikBengtsson/startup/issues
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2024-12-07 23:40:45 UTC; henrik
Author: Henrik Bengtsson [aut, cre, cph]
Maintainer: Henrik Bengtsson <henrikb@braju.com>
Repository: CRAN
Date/Publication: 2024-12-08 00:00:01 UTC
Built: R 4.6.0; ; 2026-04-27 03:56:20 UTC; unix

# Generated by roxygen2: do not edit by hand

export(check)
export(current_script)
export(find_renviron)
export(find_renviron_d)
export(find_rprofile)
export(find_rprofile_d)
export(install)
export(is_debug_on)
export(on_session_enter)
export(on_session_exit)
export(renviron_d)
export(reset_when_cache)
export(restart)
export(rprofile_d)
export(startup)
export(startup_session_options)
export(sysinfo)
export(uninstall)
export(warn)

Hello!

Welcome to the very second ALPHA release of Startup.pm

To be found at: http://wwwwbs.cs.tu-berlin.de/~schwartz/perl/

Developing with perl you certainly appreciate to write easily nifty programs. 
Unfortunately a bunch of boring problems makes life uncomfortable. With some 
of them I had to deal so often, that I created this little collection. 

Imagine:

-   You want to have an option to work on files recursively.

-   You want some modules to print some info during work, but this output 
    would hardly be predictable and thus might destroy your program output.

-   You want to have a little logfile reporting actions.

-   You want to report errors, but not via "die", and you dislike to write 
    trapping code all the time.

This module shall contribute to solve those problems, making your your 
programs more valuable.

I propose to settle Startup.pm in CPAN's category 3 as:

Startup    aupO   A program flow utility


Contact: Martin schwartz@cs.tu-berlin.de

Comments welcome! 

<div id="badges"><!-- pkgdown markup -->
<a href="https://CRAN.R-project.org/web/checks/check_results_startup.html"><img border="0" src="https://www.r-pkg.org/badges/version/startup" alt="CRAN check status"/></a> <a href="https://github.com/HenrikBengtsson/startup/actions?query=workflow%3AR-CMD-check"><img border="0" src="https://github.com/HenrikBengtsson/startup/actions/workflows/R-CMD-check.yaml/badge.svg?branch=develop" alt="R CMD check status"/></a>     <a href="https://app.codecov.io/gh/HenrikBengtsson/startup"><img border="0" src="https://codecov.io/gh/HenrikBengtsson/startup/branch/develop/graph/badge.svg" alt="Coverage Status"/></a> 
</div>

# startup: Friendly R Startup Configuration 

## Introduction

When you start R, it will by default source a `.Rprofile` file if it exists.  This allows you to automatically tweak your R settings to meet your everyday needs.  For instance, you may want to set the default CRAN repository (`options("repos")`) so you don't have to choose one every time you install a package.

The **[startup]** package extends the default R startup process by allowing you to put multiple startup scripts in a common 'Rprofile.d' directory and have them all be sourced during the R startup process.  This way you can have one file to configure the default CRAN repository and another one to configure your personal devtools settings.
Similarly, you can use a 'Renviron.d' directory with multiple files defining different environment variables.  For instance, one file may define environment variable `LANGUAGE`, whereas another file may contain your private `GITHUB_PAT` key.
The advantages of this approach are that it gives a better overview when you list the files, makes it easier to share certain settings (= certain files) with other users, and enable you to keep specific files completely private (by setting the file privileges so only you can access those settings).


## How the R startup process works

When R starts, the following _user-specific_ setup takes place:

1. The _first_ `.Renviron` file found on the R startup search path is processed.  The search path is (in order): (i) `Sys.getenv("R_ENVIRON_USER")`, (ii) `./.Renviron`, and (iii) `~/.Renviron`.  The format of this file is one `ENV_VAR=VALUE` statement per line, cf. `?.Renviron`.  _NOTE:_ Some environment variables must be set already in this step in order to be acknowledged by R, i.e. it is _too late_ to set some of them in Step 2 and Step 3 below.

2. The _first_ `.Rprofile` file found on the R startup search path is processed.  The search path is (in order): (i) `Sys.getenv("R_PROFILE_USER")`, (ii) `./.Rprofile`, and (iii) `~/.Rprofile`.  The format of this file must be a valid R script (with a trailing newline), cf. `?.Rprofile`.

3. If the `.Rprofile` file (in Step 2) calls `startup::startup()` then the following will also take place:

   a. The _first_ 'Renviron.d' directory on the R startup search path is processed.  The search path is (in order): (i) `paste0(Sys.getenv("R_ENVIRON_USER"), ".d")`, (ii) `./.Renviron.d`, (iii) `~/.Renviron.d`, and (iv) `{user-config-dir}/Renviron.d`, where `{user-config-dir}` corresponds to `tools::R_user_dir("startup", which = "config")`, e.g. `${XDG_CONFIG_HOME}/R/startup`.  The format of these files should be the same as for `.Renviron`.  _NOTE:_ Some environment variables must be set already in Step 1 above in order to be acknowledged by R.

   b. A set of handy R options that can be used in Step 3c are set.  Their names are prefixed `startup.session.` - see `?startup::startup_session_options` for details.

   c. The _first_ 'Rprofile.d' directory found on the R startup search path is processed.  The search path is (in order): (i) `paste0(Sys.getenv("R_PROFILE_USER"), ".d")`, (ii) `./.Rprofile.d`, (iii) `~/.Rprofile.d`, and (iv) `{user-config-dir}/Rprofile.d`.  The format of these files should be the same as for `.Rprofile`, that is, they must be valid R scripts.

   d. If set, any R code in environment variable `R_STARTUP_INIT`, or R option `startup.init`, is parsed and evaluated.
   
   e. If set, any R script specified by environment variable `R_STARTUP_FILE`, or R option `startup.file`, is parsed and evaluated.
   
   f. If no errors occur above, the **[startup]** package will be unloaded, leaving no trace of itself behind, except for R options `startup.session.*` set in Step 3b - these will be erased if `startup::startup()` is called with `keep = NULL`.


All relevant files in 'Renviron.d' and 'Rprofile.d'  directories, including those found recursively in subdirectories thereof, will be processed (in lexicographic order sorted under the `C` locale).
There are no restrictions on what the file names should be (except for the ones ignored as explained below).  However, for 'Rprofile.d' files we recommend to use filename extension `*.R` to indicate that the files are regular R scripts.  For 'Renviron.d' files we recommend to use files without extensions (or `*.Renviron` for clarification).  To avoid confusions, do _not_ use an `*.R` extension for 'Renviron.d' files because they are not R script per se (as some editors may warn you about).

Files with file extensions `*.txt`, `*.md` and `*~` are ignored as well as any files named `.Rhistory`, `.RData` and `.DS_Store`.  Directories named `__MACOSX` and their content are ignored.  Files and directories with names starting with two periods (`..`) are ignored, e.g. `~/.Rprofile.d/..my-tests/`.



## Installation

After installing the **startup** packages (see instructions at the end), call

```r
startup::install()
```

once.  This will append

```r
tryCatch(startup::startup(), error=function(ex) message(".Rprofile error: ", conditionMessage(ex)))
```

to your `~/.Rprofile`.  The file will be created if missing.  This will also create directories `~/.Renviron.d/` and `~/.Rprofile.d/` if missing.  To find the location of these folder on Windows, use `normalizePath("~")` - it's often located under `C:\Users\Alice\Documents\`.


Alternatively to the above installation setup, you can just add that line to your `~/.Rprofile` file manually.  The reason for using `tryCatch(..., error = ...)` is for the case when **startup** is not installed and you try to install it, e.g. after upgrading R to a new major release.  Without `tryCatch()`, R will fail to install the **startup** package (or any other package) because the R profile startup script produces an error complaining about **startup** not being available.




## Usage

Just start R :)

To debug the startup process, use `startup::startup(debug = TRUE)` or set environment variable `R_STARTUP_DEBUG=TRUE`, e.g. on Linux you can do

```sh
$ R_STARTUP_DEBUG=TRUE R
```

On MS Windows, you can do:

```sh
$ R R_STARTUP_DEBUG=TRUE
```

This will produce time-stamped messages during startup specifying which files are included together with details on the files and what they modified.


## Conditional file and directory names

If the name of a file consists of a `<key>=<value>` specification, then that file will be included / used only if the specification is fulfilled (on the current system with the current R setup).  For instance, a file `~/.Rprofile.d/os=windows.R` will be ignored unless `startup::sysinfo()$os == "windows"`, i.e. the R session is started on a Windows system.

The following `startup::sysinfo()` keys are available for conditional inclusion of files by their path names:

* System values:
  - `dirname`     - (character) the name of the current working directory (= `basename(getwd())`)
  - `gui`         - (character) the graphical user interface (= `.Platform$GUI`)
  - `nodename`    - (character) the host name (= `Sys.info()[["nodename"]]`)
  - `machine`     - (character) the machine type (= `Sys.info()[["machine"]]`)
  - `os`          - (character) the operating system (= `.Platform$OS.type`)
  - `sysname`     - (character) the system name (= `Sys.info()[["sysname"]]`)
  - `user`        - (character) the user name (= `Sys.info()[["user"]]`)
  
* System flags:
  - `interactive` - (logical) whether running R interactively or not (= `interactive()`)
  - `quiet`       - (logical) Indicates whether one of the R command-line arguments `-q`, `--quiet`, and `--silent` was specified
  - `save`        - (logical) TRUE if R was launched with command-line argument `--save`, FALSE if launched with `--no-save`, otherwise NA
  - `ess`         - (logical) whether running R in [Emacs Speaks Statistics (ESS)] or not
  - `pqr`         - (logical) whether running [pqR] ("A Pretty Quick Version of R") or not
  - `radian`      - (logical) whether running R in [radian] (formerly known as 'rtichoke' and 'rice') or not
  - `microsoftr`  - (logical) whether running R in [Microsoft R Open] or not
  - `rapp`        - (logical) whether running R in [R.app] or not (a macOS GUI)
  - `rgui`        - (logical) whether running R in [Rgui] or not (an MS Windows GUI)
  - `rstudio`     - (logical) whether running R in [RStudio] Console or not
  - `rstudioterm` - (logical) whether running R in [RStudio] Terminal or not
  - `wine`        - (logical) whether running R on Windows via [Linux Wine] or not

* Installed packages:

  - `package`     - (character) whether a package is installed or not.  In addition to checking the availability, having `package=<name>` in the filename makes it clear that the startup file concerns settings specific to that package.

* With a specific frequency:

  - `when` - (character) specify how often the file should be processed:
    - `when=once`        - the startup file is processed only once
    - `when=hourly`      - the startup file is processed at most once per hour
    - `when=daily`       - the startup file is processed at most once per day
    - `when=weekly`      - the startup file is processed at most once per week
    - `when=fortnightly` - the startup file is processed at most once every two weeks
    - `when=monthly`     - the startup file is processed at most once per month
    
  If such a file, or its timestamp, is updated, then it will be processed the next time R is started.

* Environment variables:

  - Any further `<key>=<value>` specifications with keys matching none of the above known keys are interpreted as system environment variables and **startup** will test such conditions against their values.  _Note, if `<key>` does not correspond to a known environment variable, then the file is skipped if `<key>=<value>` is used but included if `<key>!=<value>` is used._


To condition on more than one key, separate `<key>=<value>` pairs by commas, e.g. `~/.Rprofile.d/work,interactive=TRUE,os=windows.R`.  This also works for directory names.  For instance, `~/.Rprofile.d/os=windows/work,interactive=TRUE.R` will be processed if running on Windows and in interactive mode.  Multiple packages may be specified.  For instance, `~/.Rprofile.d/package=devtools,package=future.R` will be used only if both the devtools and the future packages are installed.

It is also possible to negate a conditional filename test by using the `<key>!=<value>` specification.  For instance, `~/.Rprofile.d/package=doMC,os!=windows.R` will be processed if package `doMC` is installed and the operating system is not Windows.


### Secrets (conditionally on an environment variable)

Renviron and Rprofile startup files with a non-declared `<key>` in their file names are skipped.  A non-declared key is any key that is neither one of the above predefined keys nor a declared environment variable.

This is useful, because it allows us to keep "secrets" in private files and only load them conditionally on the value of an environment variable.  For instance, we put all our secret Renviron files under `~/.Renviron.d/private/SECRET=banana/` such that they are only loaded if environment variable `SECRET` is set to exactly `banana`.  Moreover, by making sure the directory `~/.Renviron.d/private/` is only accessible by you, then it will be harder for someone else two know what your secret "password" (`banana`) is.  On a Unix system, you can set this up as:

```sh
$ mkdir -p ~/.Renviron.d/private/SECRET=banana/
$ chmod go-rw ~/.Renviron.d/private
```

Then go ahead an put your secret Renviron files in there, e.g.

```sh
.Renviron.d/private/SECRET=banana/
  +-- amazon  ## AWS_* credential & settings
  +-- github  ## GITHUB_TOKEN + GITHUB_PAT
```

Now, only if `SECRET` is set to exactly `banana` when you launch R, the above secret files will be processed and the corresponding environment variables will be available from withing R.  For instance, they will be loaded if you do:

```r
$ SECRET=banana R
```

but not if you do:

```r
$ SECRET=badguess R
```

or if `SECRET` is unset.

_Comment:_ You can use whichever variable name you like, it does not have to be `SECRET`.  And, the "password" `banana` is obviously just an example.


## Known limitations

### Setting environment variables during startup

Renviron startup files is a convenient and cross-platform way of setting environment variables during the R startup process.  However, for some of the environment variables that R consults must be set early on in the R startup process (immediately after Step 1), because R only consults them once.  Examples(*) of environment variables that need to be set _no later than_ `.Renviron` (Step 1) are:

* `TMPDIR`, `TMP`, `TEMP` - the parent of R's temporary directory, cf. `?tempdir`
* `LC_ALL` - locale settings used by R, e.g.  cf. `?locales`
* `R_DEFAULT_PACKAGES` - default set of packages loaded when R starts, cf. `?Rscript`
* `R_LIBS_USER` - user's library path, e.g. `R_LIBS_USER=~/R/%p-library/%v` is the folder specification used by default on all platforms and and R version.  The folder must exist, otherwise it is ignored by R.  The `%p` (platform) and `%v` (version) parts are R-specific conversion specifiers, cf. `?R_LIBS_USER`
* `R_MAX_NUM_DLLS`, cf. `?dyn.load`

Any changes to these done in an 'Renviron.d' file (Step 3a), or via `Sys.setenv()` in `.Rprofile` (Step 2) or 'Rprofile.d' files (Step 3c), _will be ignored by R itself_ - despite being reflected by `Sys.getenv()`.

Furthermore, some environment variables can not even be set in `.Renviron` (Step 1) but must be set _prior_ to launching R.  This is because those variables are consulted by R very early on (prior to Step 1).  Examples(*) of environment variables that need to be set _prior to_ `.Renviron` (Step 1) are:

* `HOME` - the user's home directory
* `MKL_NUM_THREADS` and `OMP_NUM_THREADS` - the default number of threads used by OpenMP etc, cf. _R Installation and Administration_
  
These variables have to be set using methods specific to the operating system or the calling shell, e.g. in a Unix shell

```sh
$ export OMP_NUM_THREADS=1
$ R
```

or per call as

```sh
OMP_NUM_THREADS=1 R
```

(*) For further details on which environment variables R consults and what they are used for by R, see the R documentation and help, e.g. `?"environment variables"` and `?Startup`.


## Examples
Below is a list of "real-world" example files:

```
.Renviron.d/
  +-- lang
  +-- r_cmd_check
  +-- secrets=banana
 
.Rprofile.d/
  +-- interactive=TRUE/
      +-- help.start.R
      +-- misc.R
	  +-- package=fortunes.R
  +-- os=windows.R
  +-- repos.R
```

They are available as part of this package under `system.file(package = "startup")`, e.g.

```r
> f <- system.file("Rprofile.d", "repos.R", package = "startup")
> file.show(f, type = "text")

local({
  repos <- c(CRAN = "https://cloud.r-project.org")
  if (.Platform$OS.type == "windows") {
     repos["CRANextra"] <- "https://www.stats.ox.ac.uk/pub/RWin"
  }
  options(repos = c(repos, getOption("repos")))
})
```

[startup]: https://cran.r-project.org/package=startup
[Emacs Speaks Statistics (ESS)]: https://ess.r-project.org/
[Microsoft R Open]: https://github.com/microsoft/microsoft-r-open
[pqR]: http://www.pqr-project.org/
[radian]: https://github.com/randy3k/radian
[RStudio]: https://posit.co/products/open-source/rstudio/
[Linux Wine]: https://www.winehq.org/

## Installation
R package startup is available on [CRAN](https://cran.r-project.org/package=startup) and can be installed in R as:
```r
install.packages("startup")
```


### Pre-release version

To install the pre-release version that is available in Git branch `develop` on GitHub, use:
```r
remotes::install_github("HenrikBengtsson/startup", ref="develop")
```
This will install the package from source.  

<!-- pkgdown-drop-below -->

## ERROR: Don't call update.packages() during startup
## because it is likely to cause an infinite recursive
## call to update.packages()
##
## startup::startup() will detect this and generate
## an informative error.
utils::update.packages()

## WARNING: It is not safe to change the 'encoding'
## option during startup.
##
## The startup::startup() function will detect this
## and give an informative warning about this.
options(encoding = "C")

## Setup built-in HTTP daemon
## Always show the HTML help on the same port
local({
  port <- sum(c(1e4, 100) * as.double(R.version[c("major", "minor")]))
  options(help.ports = port+0:9)
})

## Try to start HTML help server
try(tools::startDynamicHelp(), silent = TRUE)

options(help_type = "html")

local({
  repos <- c(CRAN = "https://cloud.r-project.org")
  if (.Platform$OS.type == "windows") {
     repos["CRANextra"] <- "https://www.stats.ox.ac.uk/pub/RWin"
  }
  repos <- c(repos, getOption("repos"))
  options(repos = repos)
})

## ERROR: If your .Rprofile file does not have a newline
## at the last line, then R itself will simply ignore
## the file silently during the startup process.
##
## The startup::startup() function will check your
## .Rprofile files for this mistake and try to fix
## the problem iff possible.
message("This line does not have a newline")

#' Checks if running R via webR
#'
#' @return A logical
#'
#' @references
#' 1. WebR - R in the Browser, <https://docs.r-wasm.org/webr/latest/>
is_webr <- function() {
  if (!"webr" %in% loadedNamespaces()) return(FALSE)

  ## Source: https://github.com/r-wasm/webr/issues/414
  ## In webr (> 0.3.2) [2024-04-23]
  if (Sys.getenv("WEBR") == "1") return(TRUE)

  ## Fallback: https://github.com/r-wasm/webr/issues/414
  ns <- getNamespace("webr")
  if (!exists("eval_js", mode = "function", envir = ns)) return(FALSE)
  eval_js <- get("eval_js", mode = "function", envir = ns)
  eval_js("'webr' in Module") == 1L
}

#' @rdname on_session_enter
#' @export
on_session_exit <- local({
  .globalenv_finalizer <- function(env) {
    "This function was added by startup::on_session_exit()"
    "Evaluate registered functions, cf. fcns <- startup::on_session_exit()"
    for (fcn in fcns) {
      local(eval(fcn(), envir = parent.frame()))
    }
  }
  
  function(fcn = NULL, append = TRUE, replace = FALSE) {
    stopifnot(is.logical(append), length(append) == 1L, !is.na(append))
    stopifnot(is.logical(replace), length(replace) == 1L, !is.na(replace))

    if (!is.function(fcn)) {
      expr <- fcn
      fcn <- function(...) NULL
      body(fcn) <- expr
    }

    env <- environment(.globalenv_finalizer)
    
    ## Registered finalizer? (only once)
    if (!isTRUE(env[["on_session_exit"]])) {
      env <- new.env(parent = globalenv())
      env[["fcns"]] <- list()
      env[["on_session_exit"]] <- TRUE
      environment(.globalenv_finalizer) <<- env
      reg.finalizer(globalenv(), .globalenv_finalizer, onexit = TRUE)
    }

    fcns <- env[["fcns"]]
    if (is.null(fcn)) return(fcns)

    ## Replace?
    if (replace) fcns <- list()
  
    ## Append or prepend?
    if (!is.null(fcn)) {
      fcn <- list(fcn)
      fcns <- if (append) c(fcns, fcn) else c(fcn, fcns)
    }

    env[["fcns"]] <- fcns

    invisible(fcns)
  }
})

#' Options and environment variables used by the 'startup' package
#'
#' Below are environment variables and \R options that are used by the
#' \pkg{startup} package.
#' The \env{R_STARTUP_***} environment variables must be set before calling
#' the `startup::startup()` function, that is, either (i) prior to launching
#' \R or (ii) in the \file{.Renviron} file.
#'
#' @section Controls whether \pkg{startup} is used or not:
#'
#' \describe{
#'   \item{\env{R_STARTUP_DISABLE} / \option{startup.disable}:}{
#'     (logical)
#'     If `TRUE`, `startup::startup()` is fully disable such that _no_
#'     \file{.Renviron.d/} or \file{.Rprofile.d/} files are processed.
#'     _Note_: Files \file{.Renviron} and \file{.Rprofile} are still processed
#'     because these are out of control of the \pkg{startup} package.
#'     (Default: `FALSE`)
#'   }
#'
#'   \item{\env{R_STARTUP_DRYRUN} / \option{startup.dryrun}:}{
#'     (logical)
#'     Controls the default value of argument `dryrun` of [startup()].
#'     (Default: `FALSE`)
#'   }
#' }
#'
#'
#' @section Additional customization of the startup process:
#'
#' \describe{
#'   \item{\env{R_STARTUP_FILE} / \option{startup.file}:}{
#'     (R script as a character string)
#'     Optional \R script that is parsed and evaluated after
#'     \file{.Renviron.d/} and \file{.Rprofile.d/} files,
#'     and `R_STARTUP_INIT` code, have been processed, e.g.
#'     `R_STARTUP_FILE="setup.R" R --quiet`.
#'     (Default: not specified)
#'   }
#'
#'   \item{\env{R_STARTUP_INIT} / \option{startup.init}:}{
#'     (R code as a character string)
#'     Optional \R code that is parsed and evaluated after
#'     \file{.Renviron.d/} and \file{.Rprofile.d/} files,
#'     but before `R_STARTUP_FILE` code, have been processed e.g.
#'     `R_STARTUP_INIT="message('Hello world')" R --quiet`.
#'     The specified string must be parsable by [base::parse()].
#'     (Default: not specified)
#'   }
#'
#'   \item{\env{R_STARTUP_RDATA} / \option{startup.rdata}:}{
#'     (comma-separated values)
#'     Controls whether an existing \file{./.RData} file should be processed 
#'     or not.
#'     If `"remove"`, it will be skipped by automatically removing it.
#'     If `"rename"`, it will be renamed to \file{./.RData.YYYYMMDD_hhmmss}
#'     where the timestamp is the last time the file was modified.
#'     If `"prompt"`, the user is prompted whether they want to load the file
#'     or rename it. In non-interactive session, `"prompt"` will fallback to 
#'     loading the content (default). To fallback to renaming the file, use
#'     `"prompt,rename"`.
#'     Note that in contrast to `R` and `R CMD BATCH file.R`, `Rscript` does
#'     _not_ load \file{.RData} files unless command-line option `--restore`
#'     is specified.  
#'     (Default: not specified)
#'   }
#' }
#'
#'
#' @section Controls what validation checks are performed at startup:
#'
#' \describe{
#'   \item{\env{R_STARTUP_CHECK} / \option{startup.check}:}{
#'     (logical)
#'     Controls the default value of argument `check` of [startup()].
#'     (Default: `TRUE`)
#'   }
#'
#'   \item{\env{R_STARTUP_CHECK_OPTIONS_IGNORE} /
#'         \option{startup.check.options.ignore}:}{
#'     (character vector or comma-separated character string)
#'     Names of \R options that should _not_ be validated at the end of the
#'     [startup()] process.
#'     (Default: `"error"`)
#'   }
#' }
#'
#'
#' @section Settings useful for debugging and prototyping:
#'
#' \describe{
#'
#'   \item{\env{R_STARTUP_DEBUG} / \option{startup.debug}:}{
#'     (logical)
#'     Controls the default value of argument `debug` of [startup()].
#'     (Default: `FALSE`)
#'   }
#'
#'   \item{\option{startup.commandArgs}:}{
#'     (character vector)
#'     Overrides the command-line arguments that [startup()] uses, which
#'     can be useful to prototype and test alternative ways that \R might
#'     be launched.
#'     (Default: `base::commandArgs()`)
#'   }
#'
#'   \item{\env{R_STARTUP_TIME} / \option{startup.time}:}{
#'     (POSIX timestamp; character string)
#'     Overrides the current timestamp, which can be useful to prototype and
#'     test functionalities that depend on the current time, e.g. inclusion
#'     and exclusion of files based on `when=<periodicity>` tags.
#'     The specified string must be parsable by [base::as.POSIXct()].
#'     (Default: not specified)
#'   }
#' }
#'
#'
#' @aliases
#' R_STARTUP_CHECK
#' startup.check
#' R_STARTUP_CHECK_OPTIONS_IGNORE
#' startup.check.options.ignore
#' startup.commandArgs
#' R_STARTUP_DEBUG
#' startup.debug
#' R_STARTUP_DRYRUN
#' startup.dryrun
#' R_STARTUP_DISABLE
#' startup.disable
#' R_STARTUP_FILE
#' startup.file
#' R_STARTUP_INIT
#' startup.init
#' R_STARTUP_RDATA
#' startup.rdata
#' R_STARTUP_TIME
#' startup.time
#'
#' @name startup.options
NULL

#' @describeIn startup Initiate using \file{.Renviron.d/} files
#'
#' @param paths (internal) character vector of directories.
#'
#' @aliases renviron
#' @export
renviron_d <- function(sibling = FALSE, all = FALSE, unload = FALSE, skip = NA,
                       dryrun = NA, debug = NA, paths = NULL) {
  debug(debug)

  ## Skip?
  if (is.na(skip)) {
    skip <- any(c("--no-environ", "--vanilla") %in% commandArgs())
  }

  if (!skip) {
    # Load custom .Renviron.d/* files
    if (is.null(paths)) paths <- find_renviron_d(sibling = sibling, all = all)
    files <- list_d_files(paths, filter = filter_files)    
    files_apply(files, fun = readRenviron,
                dryrun = dryrun, what = "Renviron",
                debug = debug)
  }

  res <- api()
  if (unload) unload()
  invisible(res)
}

#' Information on the current R session
#'
#' @return A named list.
#'
#' @examples
#' startup::sysinfo()
#'
#' @export
sysinfo <- function() {
  ## Built-in system information (character)
  sysinfo <- as.list(Sys.info())
  sysinfo$os <- .Platform$OS.type
  sysinfo$gui <- .Platform$GUI
  sysinfo$interactive <- interactive()
  
  ## Built-in system flags (logical)
  sysinfo$jupyter <- is_jupyter()
  sysinfo$rapp <- (.Platform$GUI == "AQUA")
  sysinfo$rgui <- (.Platform$GUI == "Rgui")
  sysinfo$rstudio <- is_rstudio_console()
  sysinfo$rstudioterm <- is_rstudio_terminal()
  sysinfo$microsoftr <- is_microsoftr()
  sysinfo$ess <- is_ess()
  sysinfo$radian <- is_radian()
  ## Deprecated: Renamed rtichoke -> radian in December 2018
  sysinfo$rtichoke <- sysinfo$radian
  ## Deprecated: Renamed rice -> rtichoke in February 2018
  sysinfo$rice <- sysinfo$radian
  sysinfo$pqr <- is_pqr()
  sysinfo$vscode <- is_vscode()
  sysinfo$webr <- is_webr()
  sysinfo$wine <- is_wine()

  ## Session-specific variables
  sysinfo$dirname <- basename(getwd())
  sysinfo$quiet <- any(c("-q", "--quiet", "--silent") %in% r_cli_args())
  save <- NA
  if ("--save" %in% r_cli_args()) save <- TRUE
  if ("--no-save" %in% r_cli_args()) save <- FALSE
  sysinfo$save <- save
  
  sysinfo
}


r_cli_args <- local({
  cli_args <- NULL
  function() {
    if (is.null(cli_args)) {
      cli_args <<- setdiff(commandArgs(), commandArgs(trailingOnly = TRUE))
    }
    cli_args
  }
})

#' Check if running R on Windows using Wine
#'
#' @return A logical
#'
#' @references
#' 1. Wine Developer FAQ, How can I detect Wine?,
#'    \url{https://wiki.winehq.org/Developer_FAQ#How_can_I_detect_Wine.3F}
#' 2. Jeff Zaroyko, Detecting Wine, Wine Devel mailing list, 2008-09-29,
#'    \url{https://www.winehq.org/pipermail/wine-devel/2008-September/069387.html}
is_wine <- function() {
  if (.Platform$OS.type != "windows") return(FALSE)
  any(grepl("^WINE", names(Sys.getenv())))
}

#' Check if running a Pretty Quick Version of R (pqR)
#'
#' @return A logical
#'
#' @references
#' 1. pqR - a pretty quick version of R,
#'    \url{http://www.pqr-project.org/}
#' 2. GitHub repository for 'pqR'
#'    \url{https://github.com/radfordneal/pqR}
is_pqr <- function() {
  "pqR.base.version" %in% names(R.version)
}

message("*** current_script() ...")

res <- startup::current_script()
print(res)
stopifnot(is.character(res), length(res) == 1L, is.na(res))

startup:::current_script_pathname("foo")
res <- startup::current_script()
print(res)
stopifnot(is.character(res), length(res) == 1L, res == "foo")

startup:::current_script_pathname(NA_character_)
res <- startup::current_script()
print(res)
stopifnot(is.character(res), length(res) == 1L, is.na(res))

message("*** current_script() ... DONE")

#' Checks whether startup debug is on or not
#'
#' @return Returns `TRUE` is debug is enabled and `FALSE` otherwise.
#'
#' @details
#' The debug mode is when [startup::startup()] is called, either explicitly
#' via argument `debug` or via environment variable `R_STARTUP_DEBUG`.
#' 
#' @keywords internal
#' @export
is_debug_on <- function() debug()

debug <- local({
  status <- NA

  function(new = NA) {
    if (is.na(status)) {
      t <- as.logical(Sys.getenv("R_STARTUP_DEBUG", NA))
      t <- getOption("startup.debug", t)

      ## If neither env var nor option is specified, then
      ## look at command-line options
      if (is.na(t)) {
        args <- commandArgs()
        if (any(c("-q", "--quiet", "--silent", "--slave") %in% args)) {
          t <- FALSE
        }
        if ("--verbose" %in% args) t <- TRUE
      }

      t <- isTRUE(t)

      status <<- t
    }

    new <- as.logical(new)
    if (!is.na(new)) status <<- new
    status
  }
})

log <- function(..., collapse = "\n", timestamp = TRUE, appendLF = TRUE) {
  if (!debug()) return()
  lines <- c(...)
  if (timestamp) lines <- sprintf("%s: %s", timestamp(), lines)
  message(paste(lines, collapse = collapse), appendLF = appendLF)
  invisible()
}

logf <- function(..., collapse = "\n", timestamp = TRUE, appendLF = TRUE) {
  log(sprintf(...), collapse = collapse, timestamp = timestamp, appendLF = appendLF)
}

logp <- function(expr, ...) {
  log(utils::capture.output(print(expr)), ...)
}

timestamp <- local({
  t0 <- NULL
  function(get_t0 = FALSE) {
    if (get_t0) return(t0)
    if (is.null(t0)) {
      t0 <<- Sys.time()
    }
    dt <-  difftime(Sys.time(), t0, units = "secs")
    sprintf("%5.3fs", as.numeric(dt))
  }
})

notef <- function(..., quiet = FALSE) {
  if (!quiet) message(sprintf(...))
}

files_apply <- function(files, fun,
                        on_error = c("error", "warning", "immediate.warning",
                                     "message", "ignore"),
                        dryrun = NA,
                        what = c("Renviron", "Rprofile"),
                        debug = NA) {
  stop_if_not(is.function(fun))
  on_error <- match.arg(on_error)
  what <- match.arg(what)
  debug <- debug(debug)
  
  ## Nothing to do?
  if (length(files) == 0) return(invisible(character(0)))

  ## How should the files be processed?
  if (is.na(dryrun)) {
    dryrun <- as.logical(Sys.getenv("R_STARTUP_DRYRUN", "FALSE"))
    dryrun <- getOption("startup.dryrun", dryrun)
  }

  if (isTRUE(dryrun)) {
    call_fun <- function(pathname) NULL
  } else {
    call_fun <- function(pathname) {
      res <- tryCatch(fun(pathname), error = identity)
      if (inherits(res, "error")) {
        msg <- conditionMessage(res)
        msg <- sprintf("Failure processing startup file %s: %s",
                       squote(pathname), msg)
        if (on_error == "error") {
          stop("startup::files_apply(): ", msg, call. = FALSE)
        } else if (on_error == "warning") {
          warning("startup::files_apply(): ", msg, call. = FALSE)
        } else if (on_error == "immediate.warning") {
          warning("startup::files_apply(): ", msg, immediate. = TRUE, call. = FALSE)
        } else if (on_error == "message") {
          message("startup::files_apply(): ", msg)
        }
      }
      res
    }
  }

  logf("Processing %d %s files ...", length(files), what)
  if (what == "Renviron") {
    type <- "env"
  } else if (what == "Rprofile") {
    type <- "r"
    if (debug) {
      ## (a) Loaded and attached packages
      record_pkgs <- function() {
        res <- list()
        res$loaded <- loadedNamespaces()
        res$attached <- sort(search())
        res$attached_envs <- grep("^package:", res$attached, invert = TRUE, value = TRUE)
        res$attached <- gsub("^package:", "", grep("^package:", res$attached, value = TRUE))
        res
      }

      ## (b) Environment variables
      record_envvars <- function() {
        Sys.getenv()
      }

      ## (c) R options
      record_options <- function() {
        options()
      }

      ## (d) Global variables
      record_globals <- function() {
        as.list(globalenv(), all.names = TRUE)
      }

      ## (e) Random number generator (RNG) state
      record_rng <- function() {
        list(
          kind = RNGkind(),
          seed = globalenv()$.Random.seed
        )
      }
      
      ## (f) Current working directory
      record_pwd <- function() {
        getwd()
      }
      
      ## (g) Working directory
      record_libpath <- function() {
        .libPaths()
      }
      
      ## (h) Locale
      record_locale <- function() {
        res <- strsplit(Sys.getlocale(), split = ";", fixed = TRUE)[[1]]
        pattern <- "(.*)=(.*)"
        names <- gsub(pattern, "\\1", res)
        values <- gsub(pattern, "\\2", res)
        names(values) <- names
        values
      }
    }
  }

  for (file in files) {
    ## Get 'when=<periodicity>' declaration, if it exists
    when <- get_when(file)
    logf(" - %s", file_info(file, type = type, extra = sprintf("when=%s", when)))

    if (debug && what == "Rprofile") {
      before <- list(
        envvars = record_envvars(),
        globals = record_globals(),
        libpath = record_libpath(),
         locale = record_locale(),
        options = record_options(),
           pkgs = record_pkgs(),
            pwd = record_pwd(),
            rng = record_rng()
      )
    }

    call_fun(file)

    if (debug && what == "Rprofile") {
      after <- list(
        envvars = record_envvars(),
        globals = record_globals(),
        libpath = record_libpath(),
         locale = record_locale(),
        options = record_options(),
           pkgs = record_pkgs(),
            pwd = record_pwd(),
            rng = record_rng()
      )

      ## (a) Packages
      ## Identified added entries
      added <- mapply(after$pkgs, before$pkgs, FUN = setdiff)
      added$loaded <- setdiff(added$loaded, added$attached)
      
      ## Identified removed entries
      removed <- mapply(before$pkgs, after$pkgs, FUN = setdiff)
      removed$attached <- setdiff(removed$attached, removed$loaded)
      names(removed) <- gsub("attached", "detached", names(removed))
      names(removed) <- gsub("loaded", "unloaded", names(removed))
      
      nadded <- sapply(added, FUN = length)
      nremoved <- sapply(removed, FUN = length)

      s <- NULL
      for (kind in c("attached", "loaded")) {
        if (nadded[[kind]] > 0) {
          s <- c(s, sprintf("%s %s (%s)",   nadded[[kind]], kind, paste(sQuote(  added[[kind]]), collapse = ", ")))
        }
      }
      for (kind in c("detached", "unloaded")) {
        if (nremoved[[kind]] > 0) {
          s <- c(s, sprintf("%s %s (%s)", nremoved[[kind]], kind, paste(sQuote(removed[[kind]]), collapse = ", ")))
        }
      }
      if (length(s) > 0) {
        s <- paste(s, collapse = ", ")
        logf("           Packages: %s", s, timestamp = FALSE)
      }


      ## Search path
      s <- NULL
      kind <- "attached_envs"
      if (nadded[kind] > 0) {
        s <- c(s, sprintf("%s environment attached (%s)",   nadded[[kind]],  paste(sQuote(  added[[kind]]), collapse = ", ")))
      }
      kind <- "detached_envs"
      if (nremoved[kind] > 0) {
        s <- c(s, sprintf("%s environment detached (%s)", nremoved[[kind]], paste(sQuote(  removed[[kind]]), collapse = ", ")))
      }
      if (length(s) > 0) {
        s <- paste(s, collapse = ", ")
        logf("           Search path: %s", s, timestamp = FALSE)
      }


      ## (b) Environment variables
      s <- NULL
      set <- setdiff(names(after$envvars), names(before$envvars))
      if (length(set) > 0) {
        s <- c(s, sprintf("%s added (%s)",   length(set), paste(sQuote(set), collapse = ", ")))
      }
      set <- setdiff(names(before$envvars), names(after$envvars))
      if (length(set) > 0) {
        s <- c(s, sprintf("%s removed (%s)", length(set), paste(sQuote(set), collapse = ", ")))
      }
      common <- intersect(names(before$envvars), names(after$envvars))
      diff <- which(before$envvars[common] != after$envvars[common])
      set <- names(diff)
      if (length(set) > 0) {
        s <- c(s, sprintf("%s changed (%s)", length(set), paste(sQuote(set), collapse = ", ")))
      }
      if (length(s) > 0) {
        s <- paste(s, collapse = ", ")
        logf("           Environment variables: %s", s, timestamp = FALSE)
      }


      ## (c) R options
      s <- NULL
      set <- setdiff(names(after$options), names(before$options))
      if (length(set) > 0) {
        s <- c(s, sprintf("%s added (%s)",   length(set), paste(sQuote(set), collapse = ", ")))
      }
      set <- setdiff(names(before$options), names(after$options))
      if (length(set) > 0) {
        s <- c(s, sprintf("%s removed (%s)", length(set), paste(sQuote(set), collapse = ", ")))
      }
      common <- intersect(names(before$options), names(after$options))
      same <- mapply(before$options[common], after$options[common], FUN = identical)
      set <- names(same)[!same]
      if (length(set) > 0) {
        s <- c(s, sprintf("%s changed (%s)", length(set), paste(sQuote(set), collapse = ", ")))
      }
      if (length(s) > 0) {
        s <- paste(s, collapse = ", ")
        logf("           R options: %s", s, timestamp = FALSE)
      }


      ## (d) Locale
      s <- NULL
      set <- setdiff(names(after$locale), names(before$locale))
      if (length(set) > 0) {
        s <- c(s, sprintf("%s added (%s)",   length(set), paste(sprintf("%s=%s", set, sQuote(after$locale[set])), collapse = ", ")))
      }
      set <- setdiff(names(before$locale), names(after$locale))
      if (length(set) > 0) {
        s <- c(s, sprintf("%s removed (%s)", length(set), paste(sprintf("%s=%s", set, sQuote(before$locale[set])), collapse = ", ")))
      }
      common <- intersect(names(before$locale), names(after$locale))
      same <- mapply(before$locale[common], after$locale[common], FUN = identical)
      set <- names(same)[!same]
      if (length(set) > 0) {
        s <- c(s, sprintf("%s changed (%s)", length(set), paste(sprintf("%s=%s -> %s", set, sQuote(before$locale[set]), sQuote(after$locale[set])), collapse = ", ")))
      }
      if (length(s) > 0) {
        s <- paste(s, collapse = ", ")
        logf("           Locale: %s", s, timestamp = FALSE)
      }


      ## (e) Global variables
      s <- NULL
      set <- setdiff(names(after$globals), names(before$globals))
      if (length(set) > 0) {
        s <- c(s, sprintf("%s added (%s)",   length(set), paste(sQuote(set), collapse = ", ")))
      }
      set <- setdiff(names(before$globals), names(after$globals))
      if (length(set) > 0) {
        s <- c(s, sprintf("%s removed (%s)", length(set), paste(sQuote(set), collapse = ", ")))
      }
      common <- intersect(names(before$globals), names(after$globals))
      same <- mapply(before$globals[common], after$globals[common], FUN = identical)
      set <- names(same)[!same]
      if (length(set) > 0) {
        s <- c(s, sprintf("%s changed (%s)", length(set), paste(sQuote(set), collapse = ", ")))
      }
      if (length(s) > 0) {
        s <- paste(s, collapse = ", ")
        logf("           Global variables: %s", s, timestamp = FALSE)
      }


      ## (f) Library path
      if (!identical(after$libpath, before$libpath)) {
        logf("           Library path: (%s) -> (%s)",
          paste(sprintf("%d=%s", seq_along(before$libpath), sQuote(before$libpath)), collapse = ", "),
          paste(sprintf("%d=%s", seq_along(after$libpath), sQuote(after$libpath)), collapse = ", "),
          timestamp = FALSE
        )
      }


      ## (g) Working directory
      if (!identical(after$pwd, before$pwd)) {
        logf("           Working directory: %s -> %s",
          sQuote(before$pwd),
          sQuote(after$pwd),
          timestamp = FALSE
        )
      }


      ## (h) Random number generator (RNG) state
      if (!identical(after$rng$kind, before$rng$kind)) {
        logf("           RNG kind: (%s) -> (%s)",
          paste(before$rng$kind, collapse = ", "),
          paste(after$rng$kind, collapse = ", "),
          timestamp = FALSE
        )
      }
      if (!identical(after$rng$seed, before$rng$seed)) {
        logf("           .Random.seed: updated", timestamp = FALSE)
      }

      ## Not needed anymore
      before <- after <- NULL
    }

    if (length(when) == 1L) {
      when_cache_file <- get_when_cache_file(file, when = when)
      mark_when_file_done(when_cache_file)
    }
  }

  already_done <- attr(files, "already_done", exact = TRUE)
  n_done <- length(already_done[["file"]])
  if (n_done > 0L) {
    logf(" Skipped %d files with fulfilled 'when' statements:", n_done)
    last <- vapply(already_done[["last_processed"]], FUN = format, format = "%Y-%m-%d %H:%M:%S", FUN.VALUE = NA_character_)
    logf(sprintf(" - [SKIPPED] %s (processed on %s)", squote(already_done[["file"]]), last))
  }

  unknown_keys <- attr(files, "unknown_keys", exact = TRUE)
  if (length(unknown_keys) > 0) {
    unknown_files <- names(unknown_keys)
##    for (file in unknown_files) {
##      keys <- unknown_keys[[file]]
##      reason <- sprintf("non-declared keys: %s",
##                        paste(squote(keys), collapse = ", "))
##      logf(" - [SKIPPED] %s", file_info(file, type = type, extra = reason))
##    }
    unknown_keys <- sort(unique(unlist(unknown_keys)))
    logf("[WARNING] skipped %d files with non-declared key names (%s)", length(unknown_files), paste(squote(unknown_keys), collapse = ", "))
  }
  
  logf("Processing %d %s files ... done", length(files), what)
  if (dryrun) log("(all files were skipped because startup.dryrun = TRUE)")

  invisible(files)
}

#' @describeIn startup Initiate using \file{.Rprofile.d/} files
#'
#' @aliases rprofile
#' @export
rprofile_d <- function(sibling = FALSE, all = FALSE, check = NA,
                       unload = FALSE, skip = NA,
                       on_error = c("error", "warning", "immediate.warning",
                                    "message", "ignore"),
                       dryrun = NA, debug = NA, paths = NULL) {
  debug <- debug(debug)

  if (is.na(check)) {
    check <- as.logical(Sys.getenv("R_STARTUP_CHECK", "TRUE"))
    check <- isTRUE(getOption("startup.check", check))
  }

  ## Skip?
  if (is.na(skip)) {
    skip <- any(c("--no-init-file", "--vanilla") %in% commandArgs())
  }

  # (i) Check and fix common errors
  if (check) {
    check(all = all, fix = TRUE, debug = FALSE)
  }

  debug(debug)

  if (!skip) {
    # (ii) Source custom .Rprofile.d/* files
    if (is.null(paths)) paths <- find_rprofile_d(sibling = sibling, all = all)
    files <- list_d_files(paths, filter = filter_files)
    encoding <- getOption("encoding")
    keep_source <- getOption("keep.source", TRUE)

    source_print_eval <- function(pathname) {
      current_script_pathname(pathname)
      on.exit(current_script_pathname(NA_character_))
      source(pathname, encoding = encoding, local = FALSE, chdir = FALSE,
             print.eval = TRUE,
             keep.source = keep_source, echo = FALSE, verbose = FALSE)
    }
    
    files_apply(files, fun = source_print_eval,
                on_error = on_error, dryrun = dryrun, what = "Rprofile",
                debug = debug)
  }

  res <- api()
  if (unload) unload()
  invisible(res)
}

get_startup_time <- local({
  now <- NULL
  
  function() {
    if (is.null(now)) {
      time <- Sys.getenv("R_STARTUP_TIME", "")
      time <- getOption("startup.time", time)
      if (is.character(time)) {
        if (nzchar(time)) {
          now <<- tryCatch({
            as.POSIXct(time)
          }, error = function(ex) {
            warning("Failed to parse 'R_STARTUP_TIME' as a timestamp: ",
                    squote(time), ". The reason was: ", conditionMessage(ex))
             NULL
          })
	}
      } else if (!inherits(time, "POSIXct")) {
        stop("Option 'startup.time' is of an unknown mode: ", mode(time))
      }
      if (is.null(now)) {
        now <<- Sys.time()
      }
    }
    now
  }
})

get_when_cache_path <- function(when) {
  when <- match.arg(when, choices = known_when_keys, several.ok = TRUE)
  
  cache_path <- get_user_dir(which = "cache", create = TRUE)
  path <- file.path(cache_path, when)
  path
}

get_when_cache_file <- function(pathname, when) {
  stop_if_not(length(pathname) == 1L, is_file(pathname))
  when <- match.arg(when, choices = known_when_keys)
  
  path <- get_when_cache_path(when = when)
  if (!is_dir(path)) dir.create(path, recursive = TRUE)

  ## Poor-man's file ID
  fi <- file.info(pathname)
  fi <- lapply(fi, FUN = unclass)
  is_numeric <- unlist(lapply(fi, FUN = is.numeric), use.names = FALSE)
  fi <- fi[is_numeric]
  fi <- unlist(fi, use.names = FALSE)
  file_id <- paste(c(basename(pathname), fi), collapse = "-")
  
  when_pathname <- file.path(path, file_id)
  attr(when_pathname, "pathname") <- pathname
  attr(when_pathname, "when") <- when
  
  when_pathname
}

is_when_file_done <- function(when_pathname) {
  if (!is_file(when_pathname)) {
    return(structure(FALSE, last_processed = as.POSIXct(NA)))
  }
  when <- attr(when_pathname, "when")
  stop_if_not(length(when) == 1L, is.character(when), !is.na(when))
  
  fi <- file.info(when_pathname)
  mtime <- fi[["mtime"]]

  done <- NA
  
  if (when == "once") {
    format <- "%t"  ## Trick to produce equal output
  } else if (when == "hourly") {
    format <- "%Y-%m-%d %H"
  } else if (when == "daily") {
    format <- "%Y-%m-%d"
  } else if (when == "weekly") {
    format <- "%Y %V"
  } else if (when == "fortnightly") {
    format <- "%Y %V"
    now_time <- get_startup_time()
    last_year <- format(mtime, format = "%Y")
    last_week <- as.integer(format(mtime, format = "%V"))
    last_fortnight <- floor(last_week / 2)
    now_year <- format(now_time, format = "%Y")
    now_week <- as.integer(format(now_time, format = "%V"))
    now_fortnight <- floor(now_week / 2)
    last <- sprintf("%s %02d", last_year, last_fortnight)
    now <- sprintf("%s %02d", now_year, now_fortnight)
    done <- (last >= now)
##    R.utils::mstr(list(pathname = attr(when_pathname, "pathname"), when = when, last = last, now = now, done = done))
  } else if (when == "monthly") {
    format <- "%Y %m"
  } else {
    stop("Unknown value on argument 'when': ", squote(when))
  }

  if (is.na(done)) {
    last <- format(mtime, format = format)
    now <- format(get_startup_time(), format = format)
    done <- (last >= now)
##    R.utils::mstr(list(pathname = attr(when_pathname, "pathname"), when = when, last = last, now = now, done = done))
  }

  attr(done, "last_processed") <- mtime
  
  done
}

mark_when_file_done <- function(when_pathname) {
  pathname <- attr(when_pathname, "pathname")
  timestamp <- format(get_startup_time(), format = "%Y-%m-%d %H:%M:%OS3 %z")
  cat(file = when_pathname, pathname, "\n", timestamp, "\n", sep = "")
  when_pathname
}


get_when <- function(pathname) {
  stop_if_not(length(pathname) == 1L, is.character(pathname), !is.na(pathname))

  ## Identify files specifying this <key>=<value>
  op <- "="
  pattern <- sprintf(".*[^a-z]*(when)%s([^=,/]*).*", op)
  if (!grepl(pattern, pathname, fixed = FALSE)) return(character(0L))

  pathname <- gsub("[.](r|R)$", "", pathname)
  file <- unlist(strsplit(pathname, split = "[,/]", fixed = FALSE), use.names = FALSE)
  
  when <- gsub(pattern, "\\2", file)

  ## Keep unique 'when' conditions
  when <- unique(when)
  
  ## Drop unknown 'when' conditions
  when <- intersect(when, known_when_keys)

  when
}


#' Reset all or parts of the "when" cache
#'
#' @param when (character vector) Specifies for which "when" frequencies
#' the cache should be reset.  The default is to reset all of them.
#'
#' @return (invisible) The pathnames of the "when" cache files removed.
#'
#' @export
reset_when_cache <- local({
  ## Backport
  if (getRversion() < "3.0.0") dir <- function(..., no..=TRUE) base::dir(...)
  
  function(when = c("once", "hourly", "daily", "weekly", "fortnightly", "monthly")) {
    paths <- get_when_cache_path(when = when)
    exists <- vapply(paths, FUN = is_dir, FUN.VALUE = FALSE)
    paths <- paths[exists]
    pathnames <- dir(paths, full.names = TRUE, all.files = TRUE, include.dirs = TRUE, no.. = TRUE)
    for (path in paths) unlink(path, recursive = TRUE)
    invisible(pathnames)
  }
})

known_when_keys <- eval(formals(reset_when_cache)[["when"]])

#' Check for and fix common mistakes in .Rprofile
#'
#' Check for and fix common mistakes in \file{.Rprofile} files.
#;
#' @param all Should all or only the first entry on
#' [the R startup search path][base::Startup] be checked?
#'
#' @param fix If `TRUE`, detected issues will be tried to be automatically
#' fixed, otherwise not.
#'
#' @param backup If `TRUE`, a timestamped backup copy of the original file is
#' created before modifying it, otherwise not.
#'
#' @param debug If `TRUE`, debug messages are outputted, otherwise not.
#'
#' @return Returns invisibly a character vector of files that were "fixed"
#' (modified), if any.  If no files needed to be fixed, or `fix = TRUE`,
#' then an empty vector is returned.
#' 
#' @references
#' 1. R-devel thread 'Last line in .Rprofile must have newline (PR#4056)',
#'    2003-09-03,
#'    \url{https://stat.ethz.ch/pipermail/r-devel/2003-September/027457.html}
#'
#' @export
check <- function(all = FALSE, fix = TRUE, backup = TRUE, debug = FALSE) {
  debug(debug)
  
  updated <- check_rprofile_eof(all = all, fix = fix, backup = backup,
                                debug = debug)
  
  check_rprofile_update_packages(all = all, debug = debug)
  
  if (!fix) {
    log("All startup files checked. If there were files with issues, they were not corrected because fix = FALSE.")
  } else {
    if (length(updated) == 0L) {
      log("All startup files checked. No files were fixed.")
    } else {
      logf("All startup files checked. The following files were fixed (modified): %s", paste(squote(updated), collapse = ", "))
    }
  }
  
  invisible(updated)
}


check_rprofile_eof <- function(files = NULL, all = FALSE, fix = TRUE,
                               backup = TRUE, debug = FALSE) {
  updated <- character(0L)

  debug(debug)
  if (is.null(files)) files <- find_rprofile(all = all)

  for (kk in seq_along(files)) {
    file <- files[kk]
    if (!eof_ok(file)) {
      if (fix) {
        if (backup) backup(file)

        ## Try to fix it by appending a newline
        try(cat(file = file, "\n", append = TRUE))

        ## Record that the file was updated
        updated <- c(updated, file)
        
        if (eof_ok(file)) {
          msg <- sprintf("SYNTAX ISSUE FIXED: Added missing newline to the end of file %s, which otherwise would cause R to silently ignore the file in the startup process.", file)  #nolint
          warning("startup::check(): ", msg)
        } else {
          msg <- sprintf("SYNTAX ERROR: Tried to add missing newline to the end of file %s, which otherwise would cause R to silently ignore the file in the startup process, but failed.", file)  #nolint
          stop("startup::check(): ", msg)
        }
      } else {
        msg <- sprintf("SYNTAX ERROR: File %s is missing a newline at the end of the file, which most likely will cause R to silently ignore the file in the startup process.", file)  #nolint
        stop("startup::check(): ", msg)
      }
    }
  }

  invisible(updated)
}


check_rprofile_update_packages <- function(files = NULL, all = FALSE,
                                           debug = FALSE) {
  if (is.null(files)) {
    files <- find_rprofile(all = all)
    paths <- find_rprofile_d(all = all)
    files <- c(files, list_d_files(paths))
  }
  if (length(files) == 0) return()

  patterns <- c(
    "utils::update.packages()" = "update[.]packages[(][^)]*[)]",
    "pacman::p_up()" = "p_up[(][^)]*[)]"
  )

  for (file in files) {
    bfr <- readLines(file, warn = FALSE)
    bfr <- gsub("#.*", "", bfr, fixed = FALSE)

    for (name in names(patterns)) {
      pattern <- patterns[name]
      if (any(grepl(pattern, bfr, fixed = FALSE))) {
        msg <- sprintf("UNSAFE STARTUP CALL DETECTED (%s): Updating or installing R packages during R startup will recursively spawn off an infinite number of R processes. Please remove offending call in order for .Rprofile scripts to be applied: %s", name, file)  #nolint
        stop("startup::check(): ", msg)
      }
    }
  }
}


check_options <- function(include = c("encoding", "error", "stringsAsFactors"), exclude = NA) {
  include <- match.arg(include, several.ok = TRUE,
                       choices = c("encoding", "error", "stringsAsFactors"))
  if (length(include) == 0L) return()
  if (length(exclude) > 0L) {
    if (is.na(exclude)) {
      ignore <- Sys.getenv("R_STARTUP_CHECK_OPTIONS_IGNORE", "error")
      if (is.na(ignore)) ignore <- NULL
      exclude <- getOption("startup.check.options.ignore", ignore)
    }
    keep <- (match(include, table = exclude, nomatch = 0L) == 0L)
    include <- include[keep]
  }
  if (length(include) == 0L) return()

  msg <- function(opt, default, value, body = NULL) {
    msg <- sprintf("R option '%s' was changed (to '%s') during startup, cf. Startup.  Values other than the default '%s' is known to cause problems.", opt, value, default)
    msg <- c(msg, body)
    msg <- c(msg, sprintf("To disable this check, add \"%s\" to option 'startup.check.options.ignore'.", opt))
    paste("startup::check():", paste(msg, collapse = " "))
  }

  for (opt in include) {
    if (opt == "encoding") {
      value <- getOption(opt, default = NULL)
      default <- "native.enc"
      if (!interactive() && !is.null(value) && value != default) {
        unique_warning(msg(opt, default, value, body = "For example, in non-interactive sessions installation of packages with non-ASCII characters (also in source code comments) fails. To set the encoding only in interactive mode, e.g. if (base::interactive()) options(encoding = \"UTF-8\")."), call. = FALSE)
      }
    } else if (opt == "error") {
      check_rstudio_option_error_conflict()
    } else if (opt == "stringsAsFactors") {
      value <- getOption(opt, default = NULL)
      default <- if (getRversion() >= "4.0.0") FALSE else TRUE
      if (!is.null(value) && value != default) {
        unique_warning(msg(opt, default, value), call. = FALSE)
      }
    }
  }
}


check_envs <- function() {
  check_r_libs_env_vars()
}


check_r_libs_env_vars <- function() {
  vars <- c("R_LIBS", "R_LIBS_SITE", "R_LIBS_USER")
  for (var in vars) {
    path <- Sys.getenv(var)
    if (!nzchar(path)) next

    ## Ignore "NULL" as is the case in R 4.2.0?
    if (var != "R_LIBS") {
      if (path == "NULL") next
      
      ## SPECIAL CASE: The system Renviron file sets R_LIBS_USER="%U"
      ## and R_LIBS_SITE="%S", if not already set.  Then the system
      ## Rprofile file, expands and updates their values. It keeps
      ## the values regardless of them refering to existing folders.
      ## We don't want to warn about these non-existing defaults.

      if (var == "R_LIBS_USE") {
        if (path == .expand_R_libs_env_var("%U")) {
          if (!is_dir(path)) next
        }
      } else if (var == "R_LIBS_SITE") {
        if (path == .expand_R_libs_env_var("%S") &&
	    length(.Library.site) == 0) {
          if (!is_dir(path)) next
        }
      }
    }

    ## Don't check intential "dummy" specification, e.g.
    ## non-existing-dummy-folder
    is_dummy <- grepl("^[.]", path) && !grepl("[/\\]", path)
    if (is_dummy) next

    paths <- unlist(strsplit(path, split = .Platform$path.sep, fixed = TRUE))
    paths <- unique(paths)
    paths <- paths[!vapply(paths, FUN = is_dir, FUN.VALUE = FALSE)]

    npaths <- length(paths)
    if (npaths > 0) {
      pathsx <- normalizePath(paths, mustWork = FALSE)
      pathsq <- paste(squote(paths), collapse = ", ")
      if (!all(pathsx == paths)) {
        pathsq <- sprintf("%s (expands to %s)",
                          pathsq, paste(squote(pathsx), collapse = ", "))
      }
      pathsQ <- paste(sprintf("\"%s\"", paths), collapse = ", ")
      if (npaths == 1L) {
        msg <- sprintf("Environment variable %s specifies a non-existing folder %s which R ignores and therefore are not used in .libPaths(). To create this folder, call dir.create(%s, recursive = TRUE)", squote(var), pathsq, pathsQ)
      } else {
        msg <- sprintf("Environment variable %s specifies %d non-existing folders %s which R ignores and therefore are not used in .libPaths(). To create these folders, call sapply(c(%s), dir.create, recursive = TRUE)", squote(var), npaths, pathsq, pathsQ)
      }
      unique_warning("startup::check(): ", msg, call. = FALSE)
    }
  }

  vars <- c("R_PROFILE", "R_PROFILE_USER", "R_ENVIRON", "R_ENVIRON_USER")
  for (var in vars) {
    pathname <- Sys.getenv(var)
    if (!nzchar(pathname)) next
    
    if (!is_file(pathname)) {
      pathnameq <- squote(pathname)
      pathnamex <- normalizePath(pathname, mustWork = FALSE)
      if (pathnamex != pathname) {
        pathnameq <- sprintf("%s (expands to %s)", pathnameq, squote(pathnamex))
      }
      msg <- sprintf("Environment variable %s specifies a non-existing startup file %s which R will silently ignore", squote(var), pathnameq)
      unique_warning("startup::check(): ", msg, call. = FALSE)
    }
  }

  vars <- c(build = "R_BUILD_ENVIRON", check = "R_CHECK_ENVIRON")
  for (key in names(vars)) {
    var <- vars[key]
    pathname <- Sys.getenv(var)
    if (!nzchar(pathname)) next
    
    if (!is_file(pathname)) {
      pathnameq <- squote(pathname)
      pathnamex <- normalizePath(pathname, mustWork = FALSE)
      if (pathnamex != pathname) {
        pathnameq <- sprintf("%s (expands to %s)", pathnameq, squote(pathnamex))
      }
      msg <- sprintf("Environment variable %s specifies a non-existing startup file %s which 'R CMD %s' will silently ignore", squote(var), pathnameq, key)
      unique_warning("startup::check(): ", msg, call. = FALSE)
    }
  }
}


check_rstudio_option_error_conflict <- function() {
  ## Nothing to do?
  if (is.null(getOption("error")) || !is_rstudio_console()) return()

  ## Skip check if renv is active (because we cannot reliably test)
  ## https://github.com/HenrikBengtsson/startup/issues/76
  is_renv <- isTRUE(as.logical(Sys.getenv("RENV_R_INITIALIZING")))
  if (is_renv) return()

  ## If possible, detect when 'Debug -> On Error' is _not_ set in RStudio.
  ## If so, then skip the warning, because that is a case when RStudio Console
  ## does not override 'error'.
  config_root <- "~/.rstudio-desktop"
  if (!is_dir(config_root) && sysinfo()$os == "windows") {
    ## Officially documented root folder for RStudio configuration files
    ## Source: https://support.rstudio.com/hc/en-us/articles/200534577-Resetting-RStudio-Desktop-s-State
    
    ## Alternatives on (a) Windows Vista 7, 8, ... and (b) Windows XP
    config_root <- file.path(Sys.getenv("localappdata"), "RStudio-Desktop")
    if (!is_dir(config_root)) {
      config_root <- file.path(Sys.getenv("USERPROFILE"), "Local Settings",
                               "Application Data", "RStudio-Desktop")
    }
  }
  if (is_dir(config_root)) {
    ## Non-official configuration file found by reverse engineering only,
    ## cf. https://github.com/HenrikBengtsson/startup/issues/59
    config_file <- file.path(config_root, "monitored", "user-settings",
                             "user-settings")
    if (is_file(config_file)) {
      config <- readLines(config_file, warn = FALSE)
      ## 'Debug -> On Error' is _not_ set.  Nothing to warn about
      config <- grep("errorHandlerType=", config, fixed = TRUE, value = TRUE)
      if (length(config) == 0L) return()
      if (any(grepl("errorHandlerType=\"3\"", config, fixed = TRUE))) return()
    }
  }

  ## Record intended value of option 'error'
  options(startup.error.lost = getOption("error"))

  unique_warning("startup::check(): ", "CONFLICT: Option ", squote("error"), " was set during the R startup, but this will be overridden due to the RStudio settings (menu ", squote("Debug -> On Error"), ") when using the RStudio Console. To silence this warning, do not set option 'error' when running RStudio Console, e.g. ", squote("if (!startup::sysinfo()$rstudio) options(error = ...)"), ". The 'error' option that was set during the startup process but lost is recorded in option ", squote("startup.error.lost"), ". For further details on this issue, see https://github.com/rstudio/rstudio/issues/3007")
}


## Check that Renviron and Rprofile files are properly capitalized. The proper
## way is .Renviron and .Rprofile, whereas, for instance, .REnviron is not.
warn_file_capitalization <- function(pathname, what) {
  ## Get the actual name on file
  path <- dirname(pathname)
  file <- basename(pathname)
  pattern <- sprintf("^%s$", file)
  actual <- dir(path = path, pattern = pattern, ignore.case = TRUE, all.files = TRUE)
  if (length(actual) == 0) return(invisible(TRUE))
  if (length(actual) > 1) {
    warning(sprintf("startup::startup(): Unexpected, internal result. Please report to the %s maintainer. Found more than one %s file: %s", squote(.packageName), squote(what), paste(squote(actual), collapse = ", ")))
    actual <- actual[1]
  }
  
  ## Is it a non-standard file name?
  pattern <- sprintf("^[.]?%s", what)
  if (grepl(pattern, actual)) return(invisible(TRUE))

  ## Produce informative warning
  pathname_actual <- file.path(path, actual)
  correct <- gsub(pattern, sprintf(".%s", what), actual, ignore.case = TRUE)
  pathname_correct <- file.path(path, correct)
  msg <- sprintf("Detected non-standard, platform-dependent letter casing of an %s file. Please rename file to use the officially supported casing: %s -> %s", squote(what), squote(pathname_actual), squote(pathname_correct))
  warning("startup::startup(): ", msg, call. = FALSE)
  invisible(FALSE)
}



unique_warning <- local({
  msgs <- NULL
  function(...,  call. = TRUE) {
    msg <- .makeMessage(...)
    ## Nothing to do? Already warned?
    if (msg %in% msgs) return(invisible(msg))
    msgs <<- c(msgs, msg)
    warning(..., call. = call.)
  }
})

#' Produce a warning
#'
#' @param ...,domain The message and optionally the domain used for
#' translation.  The \ldots arguments are passed to [base::sprintf] to
#' create the message string.
#'
#' @param call. If [base::TRUE], the call is included in the warning
#' message, otherwise not.
#'
#' @param immediate. If [base::TRUE], the warning is outputted immediately,
#' otherwise not.
#'
#' @export
warn <- function(..., call. = FALSE, immediate. = TRUE, domain = NULL) {
  msg <- .makeMessage(sprintf(...), domain = domain)
  files <- unlist(find_source_traceback(), use.names = FALSE)
  if (length(files) > 0) {
    files <- paste(squote(files), collapse = "->")
    msg <- sprintf("%s: %s", files, msg)
  }
 # calls <- sys.calls()
 # utils::str(calls)
  warning(msg, call. = call., immediate. = immediate., domain = NULL)
}

message("*** backup() ...")

pathname <- tempfile()
cat("Hello", file = pathname)
stopifnot(file.exists(pathname))
pathname_backup <- startup:::backup(pathname)
print(pathname_backup)
stopifnot(file.exists(pathname), file.exists(pathname_backup))

message("*** backup() ... DONE")

stop_if_not <- function(...) {
  res <- list(...)
  n <- length(res)
  if (n == 0L) return()

  for (ii in 1L:n) {
    res_ii <- .subset2(res, ii)
    if (length(res_ii) != 1L || is.na(res_ii) || !res_ii) {
        mc <- match.call()
        call <- deparse(mc[[ii + 1]], width.cutoff = 60L)
        if (length(call) > 1L) call <- paste(call[1L], "...")
        stop(sQuote(call), " is not TRUE", call. = FALSE, domain = NA)
    }
  }
}

eof_ok <- function(file) {
  size <- file_size(file)
  ## On Windows, symbolic links, e.g. by file.symlink(), give size = 0
  ## although below readBin() return a non-empty vector
  if (.Platform$OS.type == "windows" && size == 0L) size <- 1e9
  bfr <- readBin(file, what = "raw", n = size)
  n <- length(bfr)
  if (n == 0L) return(TRUE)
  is.element(bfr[n], charToRaw("\n\r"))
}

is_dir <- function(f) {
  if (length(f) != 1L) {
    stop(sprintf("INTERNAL ERROR in startup:::is_dir(): only scalar input is supported: [n=%d] %s", length(f), paste(squote(f), collapse = ", ")))
  }
  nzchar(f) && file.exists(f) && file.info(f)$isdir
}

is_file <- function(f) {
  if (length(f) != 1L) {
    stop(sprintf("INTERNAL ERROR in startup:::is_file(): only scalar input is supported: [n=%d] %s", length(f), paste(squote(f), collapse = ", ")))
  }
  nzchar(f) && file.exists(f) && !file.info(f)$isdir
}

nlines <- function(f) {
  oopts <- options(encoding = "native.enc")
  on.exit(options(oopts))
  bfr <- readLines(f, warn = FALSE)
  bfr <- grep("^[ \t]*#", bfr, value = TRUE, invert = TRUE)
  bfr <- grep("^[ \t]*$", bfr, value = TRUE, invert = TRUE)
  length(bfr)
}

## base::file.size() was only introduced in R 3.2.0
file_size <- function(...) file.info(..., extra_cols = FALSE)$size

path_info <- function(f, extra = NULL) {
  if (!nzchar(f)) return(squote(""))
  fx <- path.expand(f)
  if (!is.null(extra)) {
    extra <- paste("; ", extra, sep = "")
  } else {
    extra <- ""
  }

  if (!is_dir(f)) {
    return(sprintf("%s (non-existing directory%s)", squote(f), extra))
  }

  if (fx == f) {
    sprintf("%s (existing folder%s)", squote(f), extra)
  } else {
    sprintf("%s => %s (existing folder%s)", squote(f), squote(fx), extra)
  }
}


file_info <- function(f, type = "txt", extra = NULL, validate = FALSE) {
  if (!nzchar(f)) return(squote(""))
  fx <- path.expand(f)
  if (length(extra) > 0L) {
    extra <- paste("; ", extra, sep = "")
  } else {
    extra <- ""
  }
  
  if (!is_file(f)) {
    return(sprintf("%s (non-existing file%s)", squote(f), extra))
  }

  if (fx == f) {
    prefix <- squote(f)
  } else {
    prefix <- sprintf("%s => %s", squote(f), squote(fx))
  }
  if (type == "binary") {
    sprintf("%s (binary file; %d bytes%s)", prefix, file_size(f), extra)
  } else if (type == "env") {
    warn <- NULL
    vars <- names(parse_renviron(f))
    nvars <- length(vars)
    if (nvars > 0) {
      if (validate && !any(vars %in% names(Sys.getenv()))) {
        prefix <- sprintf("[WARNING: It appears that R never processed this file] %s", prefix)
      }
      vars <- sprintf(" (%s)", paste(squote(vars), collapse = ", "))
    } else {
      vars <- ""
    }
    sprintf("%s (%d lines; %d bytes%s) setting %d environment variables%s",
            prefix, nlines(f), file_size(f), extra, nvars, vars)
  } else if (type == "r") {
    sprintf("%s (%d code lines; %d bytes%s)",
            prefix, nlines(f), file_size(f), extra)
  } else {
    sprintf("%s (%d lines; %d bytes%s)",
            prefix, nlines(f), file_size(f), extra)
  }
}

parse_renviron <- function(f) {
  bfr <- readLines(f, warn = FALSE)
  bfr <- grep("^[ \t]*#", bfr, value = TRUE, invert = TRUE)
  bfr <- grep("^[ \t]*$", bfr, value = TRUE, invert = TRUE)
  bfr <- grep("=.*$", bfr, value = TRUE)
  pattern <- "^([^=]*)[ \t]*=[ \t]*(.*)$"
  bfr <- grep(pattern, bfr, value = TRUE)
  names <- gsub(pattern, "\\1", bfr)
  values <- gsub(pattern, "\\2", bfr)
  names(values) <- names
  values
}

find <- function(what, mode) {
  paths <- search()
  for (pos in seq_along(paths)) {
    if (exists(what, mode = mode, where = pos, inherits = FALSE)) {
      return(structure(pos, names = names(paths)[pos]))
    }
  }
  -1L
}

## sQuote() without fancy quotes, if possible
squote <- if (getRversion() >= "3.6.0") {
  function(x) sQuote(x, q = FALSE)
} else {
  function(x) sQuote(x)
}

supports_tcltk <- function() {
  (capabilities("X11") && capabilities("tcltk") &&
   requireNamespace("tcltk") && suppressWarnings(tcltk::.TkUp))
}

tcltk_yesno <- function(question) {
  if (!supports_tcltk()) {
    warning("[startup::startup()]: Your R setup does not support X11 or tcltk dialogs. Because of this, the answer to the question ", sQuote(question), " was defaulted to ", sQuote("yes"), ".", call. = FALSE)
    return(TRUE)
  }  
  ans <- tcltk::tk_messageBox("yesno", message = question)
  (ans == "yes")
}

ask_yes_no <- function(question, rdata_workaround = TRUE) {
  ## RStudio Console workarounds?
  if (is_rstudio_console()) {
    if (rdata_workaround) {
      ## WORKAROUND: RStudio Console will load any .RData file as soon as
      ## base::readline() or utils::menu(..., graphics = FALSE) is called
      ## during startup process (https://github.com/rstudio/rstudio/issues/5844)
      ## Instead, we use a TclTk dialog.  If this is not possible, we will
      ## produce a warning and default to 'Yes'.
      return(tcltk_yesno(question))
    }
    
    ## WORKAROUND: RStudio Console does not show the base::readline() prompt
    ## during startup process (https://github.com/rstudio/rstudio/issues/5842)
    readline <- function(prompt) {
      ## Comment: appendLF = FALSE makes no difference. The "readline"
      ## will trigger a "> " prompt to be display on the next line
      message(prompt, appendLF = FALSE)
      base::readline(prompt = "")
    }
  }

  prompt <- sprintf("%s [Y/n]: ", question)
  res <- TRUE
  repeat({
    ans <- readline(prompt)
    ans <- gsub("(^[[:space:]]*|[[:space:]]*$)", "", ans)
    ans <- tolower(ans)
    if (ans %in% c("", "y", "yes")) {
      res <- TRUE
      break
    } else if (ans %in% c("n", "no")) {
      res <- FALSE
      break
    }
  })

  res
}

list_of_values <- function(files, pattern, names = FALSE) {
  if (length(files) == 0) return(list())
  files <- gsub("[.](r|R)$", "", files)
  files <- strsplit(files, split = "[,/]", fixed = FALSE)
  files <- lapply(files, FUN = function(f) grep(pattern, f, value = TRUE))
  if (names) {
    lapply(files, FUN = function(f) {
      values <- gsub(pattern, "\\2", f)
      names(values) <- gsub(pattern, "\\1", f)
      values
    })
  } else {
    lapply(files, FUN = function(f) gsub(pattern, "\\2", f))
  }
}


filter_files_info <- function(files, info = sysinfo()) {
  for (op in c("=", "!=")) {
    ## Parse <key>=<value> and keep only matching ones
    for (key in names(info)) {
      ## Identify files specifying this <key>=<value> or <key>!=<value>
      pattern <- sprintf(".*[^a-z]*(%s)%s([^=,/]*).*", key, op)
      idxs <- grep(pattern, files, fixed = FALSE)
      if (length(idxs) == 0) next

      ## There could be more than one <key>=<name> specification
      ## per pathname that use the same <key>, e.g. package=nnn.
      files_values <- list_of_values(files[idxs], pattern = pattern)

      ## sysinfo() keys
      value <- info[[key]]
      if (is.logical(value)) {
        files_values <- unlist(files_values, use.names = FALSE)
        files_values <- toupper(files_values)
        files_values[files_values == "1"] <- "TRUE"
        files_values[files_values == "0"] <- "FALSE"
        files_values <- as.logical(files_values)
        ## SPECIAL CASE: For all but 'save', interpret NA as FALSE
        if (key != "save") {
          files_values[is.na(files_values)] <- FALSE
        }
      }
      files_ok <- lapply(files_values, FUN = function(values) {
        keep <- vapply(values, FUN = identical, value, FUN.VALUE = FALSE)
        if (op == "!=") keep <- !keep
        all(keep)
      })
      files_ok <- unlist(files_ok, use.names = FALSE)
      drop <- idxs[!files_ok]

      if (length(drop) > 0) files <- files[-drop]
    } ## for (key ...)
  } ## for (op ...)

  files
} ## filter_files_info()


filter_files_package <- function(files) {
  for (op in c("=", "!=")) {
    ## Parse <key>=<value> and keep only matching ones

    ## Identify files specifying this <key>=<value> or <key>!=<value>
    pattern <- sprintf(".*[^a-z]*(package)%s([^=,/]*).*", op)
    idxs <- grep(pattern, files, fixed = FALSE)
    if (length(idxs) == 0) next

    ## There could be more than one <key>=<name> specification
    ## per pathname that use the same <key>, e.g. package=nnn.
    files_values <- list_of_values(files[idxs], pattern = pattern)

    files_ok <- lapply(files_values, FUN = function(values) {
      ## Check which packages are installed and can be loaded
      keep <- lapply(values, FUN = is_package_installed)
      keep <- unlist(keep, use.names = FALSE)
      if (op == "!=") keep <- !keep
      all(keep)
    })
    files_ok <- unlist(files_ok, use.names = FALSE)
    drop <- idxs[!files_ok]

    if (length(drop) > 0) files <- files[-drop]
  } ## for (op ...)

  files
} ## filter_files_package()


filter_files_when <- function(files) {
  already_done <- list(
    file = NULL,
    last_processed = NULL
  )
  
  for (op in c("=")) {
    ## Parse <key>=<value> and keep only matching ones

    ## Identify files specifying this <key>=<value>
    pattern <- sprintf(".*[^a-z]*(when)%s([^=,/]*).*", op)
    idxs <- grep(pattern, files, fixed = FALSE)
    if (length(idxs) == 0) next

    ## There could be more than one <key>=<name> specification
    ## per pathname that use the same <key>, e.g. package=nnn.
    files_values <- list_of_values(files[idxs], pattern = pattern)

    ## Keep unique 'when' conditions
    files_values <- lapply(files_values, FUN = unique)

    ## Drop unknown 'when' conditions
    files_values <- lapply(files_values, FUN = function(when) {
      intersect(when, known_when_keys)
    })

    ## Ignore multiple *different* or empty 'when' statements
    n <- unlist(lapply(files_values, FUN = length), use.names = FALSE)
    if (any(n != 1L)) {
      non_unique <- files[idxs][n > 1L]
      keep <- which(n == 1L)
      idxs <- idxs[keep]
      if (length(idxs) == 0) next
      files_values <- files_values[keep]
    }

    last_processed_op <- vector("list", length = length(idxs))
    done <- logical(length(idxs))
    for (kk in seq_along(idxs)) {
      file <- files[idxs[kk]]
      when <- files_values[[kk]]
      when_file <- get_when_cache_file(file, when = when)
      done_kk <- is_when_file_done(when_file)
      done[kk] <- done_kk
      last_processed_op[kk] <- list(attr(done_kk, "last_processed"))
    }
    
    already_done[["file"]] <- c(already_done[["file"]], files[idxs][done])
    already_done[["last_processed"]] <- c(already_done[["last_processed"]], last_processed_op[done])
    
    drop <- idxs[done]

    if (length(drop) > 0) files <- files[-drop]
  } ## for (op ...)

  if (length(already_done[["file"]]) > 0L) {
    attr(files, "already_done") <- already_done
  }

  files
} ## filter_files_when()


filter_files_env <- function(files, ignore = c(names(sysinfo()), "package")) {
  envs <- Sys.getenv()

  unknown_keys <- NULL
  
  for (op in c("=", "!=")) {
    ## Identify files specifying this <key>=<value> or <key>!=<value>
    pattern <- sprintf("^([a-zA-Z_][a-zA-Z0-9_]*)%s([^=,/]*).*", op)
    files_values <- list_of_values(files, pattern = pattern, names = TRUE)

    ## Drop <key>=<value> elements that refers to sysinfo() or packages
    files_values <- lapply(files_values, FUN = function(x) {
      x[!names(x) %in% ignore]
    })

    idxs <- which(sapply(files_values, FUN = length) > 0)
    if (length(idxs) == 0) next

    ## There could be more than one <key>=<name> specification
    ## per pathname that use the same <key>, e.g. package=nnn.
    env_names <- names(envs)
    files_ok <- lapply(files_values[idxs], FUN = function(values) {
      keys <- names(values)
      
      ## Check for keys for which there exist *no* environment variable
      missing <- keys[is.na(match(keys, env_names))]
      if (length(missing) > 0) return(missing)
      
      truth <- envs[keys]
      keep <- (values == truth)
      if (op == "!=") keep <- !keep
      all(keep)
    })

    
    ## Any files with un-declared 'key' in their pathnames?
    has_unknown_keys <- which(unlist(lapply(files_ok, FUN = is.character)))
    if (length(has_unknown_keys) > 0) {
      unknown_keys_op <- files_ok[has_unknown_keys]
      names(unknown_keys_op) <- files[idxs[has_unknown_keys]]
      if (op == "=") {
        unknown_keys <- c(unknown_keys, unknown_keys_op)
        files_ok[has_unknown_keys] <- FALSE
      } else {
        files_ok[has_unknown_keys] <- TRUE
      }
    }

    files_ok <- unlist(files_ok, use.names = FALSE)
    drop <- idxs[!files_ok]

    if (length(drop) > 0) files <- files[-drop]
  } ## for (op ...)

  attr(files, "unknown_keys") <- unknown_keys

  files
} ## filter_files_env()


## base::anyNA() was introduced in R 3.1.0
if (!exists("anyNA", mode = "function", envir = baseenv(), inherits = FALSE)) {
  anyNA <- function(x) any(is.na(x))
}
  
filter_files <- function(files, info = sysinfo()) {
  files0 <- files
  
  files <- filter_files_info(files, info = info)
  
  files <- filter_files_package(files)
  
  files <- filter_files_when(files)
  already_done <- attr(files, "already_done")
  
  files <- filter_files_env(files, ignore = c(names(info), "package", "when"))

  attr(files, "already_done") <- already_done

  ## Sanity checks
  if (anyNA(files)) {
      stop("[INTERNAL ERROR] startup:::filter_files() produced missing values: ", paste(sprintf("%s -> %s", squote(files0), squote(files)), collapse = ", "))
  }

  files
} ## filter_files()


is_package_installed <- local({
  cache <- list()
  function(pkg) {
    res <- cache[[pkg]]
    if (is.logical(res)) return(res)
    res <- (length(find.package(package = pkg, lib.loc = .libPaths(),
                                quiet = TRUE)) > 0)
    cache[[pkg]] <<- res
    res
  }
})

find_source_traceback <- function() {
  # Identify the environment/frame of interest by making sure
  # it at least contains all the arguments of source().
  argsToFind <- names(formals(base::source))

  # Scan the call frames/environments backwards...
  srcfileList <- list()
  for (ff in sys.nframe():0) {
    env <- sys.frame(ff)

    # Does the environment look like a source() environment?
    exist <- sapply(argsToFind, FUN=exists, envir=env, inherits=FALSE)
    if (!all(exist)) {
      # Nope, then skip to the next one
      next
    }

    # Identity the source file
    if (exists("srcfile", envir=env, inherits=FALSE)) {
      srcfile <- get("srcfile", envir=env, inherits=FALSE)
    } else {
      ## AD HOC: Needed when for instance 'XML' is attached
      srcfile <- get("srcfile", envir=env, inherits=TRUE)
      ## Failed?
      if (!inherits(srcfile, "srcfile")) srcfile <- NULL
    }
    
    if (!is.null(srcfile)) {
      if (!is.function(srcfile)) {
        srcfileList <- c(srcfileList, list(srcfile))
      }
    }
  } # for (ff ...)

  # Extract the pathnames to the files called
  pathnames <- sapply(srcfileList, FUN=function(srcfile) {
    if (inherits(srcfile, "srcfile")) {
      pathname <- srcfile$filename
    } else if (is.environment(srcfile)) {
      pathname <- srcfile$filename
    } else if (is.character(srcfile)) {
      # Occurs with source(..., keep.source=FALSE)
      pathname <- srcfile
    } else {
      pathname <- NA_character_
      warning("Unknown class of 'srcfile': ", class(srcfile)[1L])
    }
    pathname
  })
  names(srcfileList) <- pathnames

  srcfileList
}

#' Load .Renviron.d and .Rprofile.d directories during the R startup process
#'
#' Initiates \R using all files under \file{.Renviron.d/} and / or
#' \file{.Rprofile.d/} directories (or in subdirectories thereof).
#'
#' The above is done in addition the \file{.Renviron} and \file{.Rprofile}
#' files that are supported by the built-in [startup process][base::Startup]
#' of \R.
#'
#' @param sibling If `TRUE`, then only \file{.Renviron.d/} and
#' \file{.Rprofile.d/} directories with a sibling \file{.Renviron} and
#' \file{.Rprofile} in the same location will be considered.
#'
#' @param all If `TRUE`, then _all_ \file{.Renviron.d/} and \file{.Rprofile.d/}
#' directories found on [the R startup search path][base::Startup] are
#' processed, otherwise only the _first ones_ found.
#'
#' @param on_error Action taken when an error is detected when sourcing an
#' Rprofile file.  It is not possible to detect error in Renviron files;
#' they are always ignored with a message that cannot be captured.
#'
#' @param keep Specify what information should remain after this function
#' complete.  The default is to keep `startup.session.*` options
#' as recorded by [startup_session_options()].
#'
#' @param encoding The encodingto use when parsing the R startup files.
#' See [base::parse()] for more details.
#' 
#' @param check If `TRUE`, then the content of startup files are validated.
#' 
#' @param unload If `TRUE`, then the package is unloaded afterward, otherwise
#' not.
#'
#' @param skip If `TRUE`, startup directories will be skipped.  If `NA`, they
#' will be skipped if command-line options `--vanilla`, `--no-init-file`,
#' and / or `--no-environ` were specified.
#'
#' @param dryrun If `TRUE`, everything is done except the processing of the
#' startup files.
#'
#' @param debug If `TRUE`, debug messages are outputted, otherwise not.
#'
#' @section User-specific installation:
#' In order for \file{.Rprofile.d} and \file{.Renviron.d} directories to be
#' included during the \R startup process, a user needs to add
#' `startup::startup()` to \file{~/.Rprofile}.  Adding this can also be done
#' by calling [startup::install()] once.
#'
#' @section Site-wide installation:
#' An alternative to having each user add `startup::startup()` in their own
#' \file{~/.Rprofile} file, is to add it to the site-wide \file{Rprofile.site}
#' file (see [?Startup][base::Startup]).
#' The advantage of such a site-wide installation, is that the users do not
#' have to have a \file{.Rprofile} file for \file{.Rprofile.d} and
#' \file{.Renviron.d} directories to work.
#' For this to work for all users automatically, the \pkg{startup} package
#' should also be installed in the site-wide library.
#'
#' @examples
#' \dontrun{
#' # The most common way to use the package is to add
#' # the following call to the ~/.Rprofile file.
#' startup::startup()
#'
#' # To process ~/.Renviron.d/ files, and then any ./.Renviron.d/ files,
#' # followed by  ~/.Rprofile.d/ files, and then any ./.Rprofile.d/ files,
#' # add the following call to the ~/.Rprofile file.
#' startup::startup(all = TRUE)
#'
#' # For finer control of on exactly what files are used
#' # functions renviron_d() and rprofile_d() are also available:
#'
#' # Initiate first .Renviron.d/ found on search path
#' startup::renviron_d()
#'
#' # Initiate all .Rprofile.d/ directories found on the startup search path
#' startup::rprofile_d(all = TRUE)
#' }
#'
#' @describeIn startup `renviron_d()` followed by `rprofile_d()` and then the
#' package is unloaded
#' @export
startup <- function(sibling = FALSE, all = FALSE,
                    on_error = c("error", "warning", "immediate.warning",
                                 "message", "ignore"),
                    keep = c("options"), check = NA, unload = TRUE, skip = NA,
                    encoding = getOption("encoding"),
                    dryrun = NA, debug = dryrun) {
  ## Is startup::startup() fully disabled?
  disable <- as.logical(Sys.getenv("R_STARTUP_DISABLE", "FALSE"))
  disable <- getOption("startup.disable", disable)
  if (isTRUE(disable)) {
    return(invisible())
  }

  on_error <- match.arg(on_error)
  if (length(keep) > 0) keep <- match.arg(keep, several.ok = TRUE)

  if (is.na(check)) {
    check <- as.logical(Sys.getenv("R_STARTUP_CHECK", "TRUE"))
    check <- isTRUE(getOption("startup.check", check))
  }
  
  debug(debug)

  cmd_args <- getOption("startup.commandArgs", commandArgs())

  debug <- debug()
  if (debug) {
    r_home <- R.home()
    r_arch <- .Platform$r_arch
    r_os <- .Platform$OS.type

    logf("System information:")
    logf("- R_HOME: %s", path_info(Sys.getenv("R_HOME")))
    logf("- R call: %s", paste(cmd_args, collapse = " "))
    logf("- Current directory: %s", squote(getwd()))
    logf("- User's home directory: %s", path_info("~"))
    logf("- User's %s config directory: %s", squote(.packageName), path_info(get_user_dir("config")))
    logf("- Search path: %s", paste(squote(search()), collapse = ", "))
    logf("- Loaded namespaces: %s",
         paste(squote(loadedNamespaces()), collapse = ", "))

    logf("The following has already been processed by R:")

    logf("- R_ENVIRON: %s", file_info(Sys.getenv("R_ENVIRON"), type = "env", validate = TRUE))
    logf("- R_ENVIRON_USER: %s", file_info(Sys.getenv("R_ENVIRON_USER"), type = "env", validate = TRUE))
    
    if (r_os == "unix") {
      f <- file.path(r_home, "etc", "Renviron")
      if (is_file(f)) logf("- %s", file_info(f, type = "env", validate = TRUE))
    }

    no_environ <- any(c("--no-environ", "--vanilla") %in% cmd_args)
    if (!no_environ) {
      f <- Sys.getenv("R_ENVIRON")
      if (nzchar(r_arch) && !is_file(f)) {
        f <- file.path(r_home, "etc", r_arch, "Renviron.site")
      }
      if (!is_file(f)) f <- file.path(r_home, "etc", "Renviron.site")
      if (is_file(f)) logf("- %s", file_info(f, type = "env", validate = TRUE))

      f <- Sys.getenv("R_ENVIRON_USER")
      if (is_file(f)) {
        logf("- %s", file_info(f, type = "env", validate = TRUE))
      } else {
        if (nzchar(r_arch) && !is_file(f)) f <- sprintf(".Renviron.%s", r_arch)
        f <- ".Renviron"
        if (nzchar(r_arch) && !is_file(f)) f <- sprintf("~/.Renviron.%s", r_arch)
        if (!is_file(f)) f <- "~/.Renviron"
        if (is_file(f)) {
          logf("- %s", file_info(f, type = "env", validate = TRUE))
          warn_file_capitalization(f, "Renviron")
        }
      }
    }

    ## TMPDIR et al. may be set at the latest in an Renviron file
    logf("- tempdir(): %s", path_info(tempdir()))
    for (name in c("TMPDIR", "TMP", "TEMP")) {
      value <- Sys.getenv(name, "")
      logf("  - %s: %s", name, sQuote(value))
    }

    logf("- R_LIBS: %s", squote(Sys.getenv("R_LIBS")))
    logf("- R_LIBS_SITE: %s", squote(Sys.getenv("R_LIBS_SITE")))
    logf("- R_LIBS_USER: %s", squote(Sys.getenv("R_LIBS_USER")))

    pkgs <- Sys.getenv("R_SCRIPT_DEFAULT_PACKAGES")
    logf("- R_SCRIPT_DEFAULT_PACKAGES (only if Rscript was used): %s", squote(pkgs))
    
    pkgs <- Sys.getenv("R_DEFAULT_PACKAGES")
    if (pkgs == "") {
      ## In R (< 3.5.0), the 'methods' package is _not_ attached when Rscript
      ## is used.  In R (>= 3.5.0), the 'methods' package is always attached
      ## by default.  If attached, the 'methods' package is attached at the
      ## very beginning when R is started moments after the 'base' package is
      ## attached.  This is contrary to all other packages which are attached
      ## below.
      if (getRversion() < "3.5.0" && is_rscript()) {
        pkgs <- "base,datasets,utils,grDevices,graphics,stats"
      } else {
        pkgs <- "base,methods,datasets,utils,grDevices,graphics,stats"
      }
      logf("- R_DEFAULT_PACKAGES: %s (= %s)", squote(""), squote(pkgs))
    } else {
      logf("- R_DEFAULT_PACKAGES: %s", squote(pkgs))
    }

    f <- system.file("R", "Rprofile", package = "base")
    if (is_file(f)) logf("- %s", file_info(f, type = "r"))

    logf("- R_PROFILE: %s", file_info(Sys.getenv("R_PROFILE")))
    no_site_file <- any(c("--no-site-file", "--vanilla") %in% cmd_args)
    if (!no_site_file) {
      f <- Sys.getenv("R_PROFILE")
      if (is_file(f)) {
        logf("- %s", file_info(f, type = "r", validate = TRUE))
      } else {
        if (nzchar(r_arch)) f <- file.path(r_home, "etc", r_arch, "Rprofile.site")
        if (!is_file(f)) f <- file.path(r_home, "etc", "Rprofile.site")
        if (is_file(f)) {
          logf("- %s", file_info(f, type = "r"))
          warn_file_capitalization(f, "Rprofile")
        }
      }
    }

    logf("- R_PROFILE_USER: %s", file_info(Sys.getenv("R_PROFILE_USER")))
    no_init_file <- any(c("--no-init-file", "--vanilla") %in% cmd_args)
    if (!no_init_file) {
      f <- Sys.getenv("R_PROFILE_USER")
      if (is_file(f)) {
        logf("- %s", file_info(f, type = "r", validate = TRUE))
      } else {
        if (nzchar(r_arch)) f <- sprintf(".Rprofile.%s", r_arch)
        if (!is_file(f)) f <- ".Rprofile"
        if (nzchar(r_arch) && !is_file(f)) f <- sprintf("~/.Rprofile.%s", r_arch)
        if (!is_file(f)) f <- "~/.Rprofile"
        if (is_file(f)) {
          logf("- %s", file_info(f, type = "r"))
          warn_file_capitalization(f, "Rprofile")
        }
      }
    }

    f <- Sys.getenv("R_TESTS")
    if (nzchar(f)) {
      if (is_file(f)) {
        logf("- R_TESTS: %s", file_info(f, type = "r"))
      } else {
        logf("- R_TESTS: %s (not found)", normalizePath(f, mustWork = FALSE))
      }
    }

    if (.Platform$GUI == "Rgui") {
      log("Rgui configuration files (MS Windows only):")
      paths <- c(user = Sys.getenv("R_USER"), system = file.path(Sys.getenv("R_HOME"), "etc"))
      already_done <- FALSE
      for (name in names(paths)) {
        path <- paths[[name]]
        if (!is_dir(path)) next
        f <- file.path(path, "Rconsole")
        if (is_file(f)) {
          if (already_done) {
            info <- file_info(f, type = "r")
            info <- sprintf("%s [skipped]", info)
          } else {
            info <- file_info(f, type = "r")
            already_done <- TRUE
          }
        } else {
          info <- sprintf("%s (not found)", normalizePath(f, mustWork = FALSE))
        }
        logf("- %s-specific Rconsole configuration: %s", name, info)
      } ## for (path ...)
    }
  }

  logf("startup::startup()-specific processing ...")

  # (i) Load custom .Renviron.d/* files
  renviron_d(sibling = sibling, all = all, skip = skip, dryrun = dryrun)

  # (ii) Load custom .Rprofile.d/* files
  rprofile_d(sibling = sibling, all = all, check = check, skip = skip,
             dryrun = dryrun, on_error = on_error)


  ## (iv) Detect and report on run-time startup issues
  if (check) {
    # (a) Check for unsafe/non-intended changes to environment variables
    #     to library, Renviron, or Rprofile paths
    check_envs()
    
    # (b) Check for unsafe changes to R options changes done by
    #     any Rprofile files
    check_options()
  }

  ## (iii) Process R_STARTUP_INIT code?
  code <- Sys.getenv("R_STARTUP_INIT")
  code <- getOption("startup.init", code)
  if (nzchar(code)) {
    logf("Processing R_STARTUP_INIT/startup.init=%s:", squote(code))
    expr <- tryCatch(parse(text = code, encoding = encoding), error = identity)
    if (inherits(expr, "error")) {
      msg <- sprintf("Syntax error in 'R_STARTUP_INIT'/'startup.init': %s", squote(code))
      logf(paste("- [SKIPPED]", msg))
      if (on_error == "error") {
        stop(msg, call. = FALSE)
      } else if (on_error == "warning") {
        warning(msg, call. = FALSE)
      } else if (on_error == "immediate.warning") {
        warning(msg, immediate. = TRUE, call. = FALSE)
      } else if (on_error == "message") {
        message(msg)
      }
    } else {
      eval(expr, envir = .GlobalEnv, enclos = baseenv())
    }
  }

  ## (iv) Detect and report on run-time startup issues
  if (check) {
    check_envs()
    check_options()
  }
  
  ## (iii) Process R_STARTUP_FILE code?
  f <- Sys.getenv("R_STARTUP_FILE")
  f <- getOption("startup.file", f)
  if (nzchar(f)) {
    logf("Processing R_STARTUP_FILE/startup.file=%s:", squote(f))
    if (!is_file(f)) {
      msg <- sprintf("No such file 'R_STARTUP_FILE'/'startup.file' file: %s", squote(f))
      logf(paste("- [SKIPPING]", msg))
      if (on_error == "error") {
        stop(msg, call. = FALSE)
      } else if (on_error == "warning") {
        warning(msg, call. = FALSE)
      } else if (on_error == "immediate.warning") {
        warning(msg, immediate. = TRUE, call. = FALSE)
      } else if (on_error == "message") {
        message(msg)
      }
    }
    expr <- tryCatch(parse(file = f, encoding = encoding), error = identity)
    if (inherits(expr, "error")) {
      msg <- sprintf("Syntax error in 'R_STARTUP_INIT'/'startup.init': %s", squote(code))
      logf(paste("- [SKIPPED]", msg))
      if (on_error == "error") {
        stop(msg, call. = FALSE)
      } else if (on_error == "warning") {
        warning(msg, call. = FALSE)
      } else if (on_error == "immediate.warning") {
        warning(msg, immediate. = TRUE, call. = FALSE)
      } else if (on_error == "message") {
        message(msg)
      }
    } else {
      eval(expr, envir = .GlobalEnv, enclos = baseenv())
    }
  }

  ## (iv) Detect and report on run-time startup issues
  if (check) {
    check_envs()
    check_options()
  }
  
  res <- api()

  ## (v) Cleanup?
  if (!"options" %in% keep) startup_session_options(action = "erase")

  # (vi) Unload package?
  if (unload) {
    if (debug) logf("- unloading the %s package", squote(.packageName))
    on.exit(unload(debug = FALSE))
  }

  if (debug) {
    values <- search()
    if (unload) values <- setdiff(values, sprintf("package:%s", .packageName))
    logf("- Search path: %s", paste(squote(values), collapse = ", "))
    values <- loadedNamespaces()
    if (unload) values <- setdiff(values, .packageName)
    logf("- Loaded namespaces: %s", paste(squote(values), collapse = ", "))
    logf("startup::startup()-specific processing ... done")
    logf("The following will be processed next by R:")
  }

  no_restore_data <- any(c("--no-restore-data", "--no-restore", "--vanilla") %in% cmd_args)
  loads_RData <- has_RData <- FALSE
  if (!no_restore_data) {
    has_RData <- is_file(f <- "./.RData")
    if (has_RData) {
      f_norm <- normalizePath(f)
      f_info_short <- file_info(f, type = "binary")
      f_info <- file_info(f_norm, type = "binary")
      rdata <- Sys.getenv("R_STARTUP_RDATA", "")
      rdata <- getOption("startup.rdata", rdata)
      rdata0 <- rdata
      if (length(rdata) == 0L || rdata == "") {
        rdata <- "default"
      } else if (debug) {
        logf("- R_STARTUP_RDATA/startup.rdata=%s", paste(rdata, collapse = ","))
      }

      ## Support R_STARTUP_RDATA/startup.rdata=prompt,rename
      rdata <- unlist(strsplit(rdata, split = ",", fixed = TRUE))
      stopifnot(length(rdata) >= 1L, length(rdata) <= 2L)
      
      if (rdata[1] == "prompt") {
        fallback <- rdata[2L]
        if (interactive()) {  
          if (is.na(fallback) || fallback == "default") fallback <- "rename"
          if (debug) logf("- Prompting user whether they want to load or %s %s", fallback, f_info)
          question <- sprintf("Detected %s - do you want to load it? If not, it will be %sd.", f_info, fallback)

          ## We might be able to prompt the user
          if (is_rstudio_console() && !supports_tcltk()) {
            rdata <- "default"
            if (debug) logf("- Cannot prompt user in the RStudio Console on this system")
            warning(sprintf("Detected %s, which was loaded (default), because it was possible to ask you if it should loaded or not. The reason for this is that your R setup does not support X11 or tcltk, which is needed in order to prompt someone in the RStudio Console.", f_info, rdata0), call. = FALSE)
          } else {
            res <- ask_yes_no(question)
            if (debug) logf("- User wants to load it: %s", res)
            rdata <- if (res) "default" else fallback
          }  
        } else {
          ## Non-interactive session; it is not possible to the prompt user.
          if (length(rdata) == 1L) {
            rdata <- "default"
            warning(sprintf("Loading %s because it is not possible to prompt the user in a non-interactive session [R_STARTUP_RDATA/startup.rdata=%s]", f_info, rdata0), call. = FALSE)
          } else {
            ## Use fallback
            stop_if_not(!is.na(fallback))
            rdata <- fallback
          }
        }
      }

      ## At this point, we should have at most one element in 'rdata'
      stop_if_not(length(rdata) == 1L, !is.na(rdata))
      
      if (rdata == "remove") {
        if (debug) logf("- Skipping %s by removing it", f_info)
        file.remove(f)
        has_RData <- is_file(f)
        if (!has_RData) {
          warning(sprintf("Skipped %s by removing it [R_STARTUP_RDATA/startup.rdata=%s]", f_info, rdata0), call. = FALSE)
        }
      } else if (rdata == "rename") {
        fi <- file.info(f)
        when <- fi[c("mtime", "ctime")]
        keep <- vapply(when, FUN = inherits, "POSIXct", FUN.VALUE = FALSE)
        when <- when[keep]
        when <- unlist(when, use.names = FALSE)
        when <- sort(when, decreasing = TRUE)
        when <- format(when[[1]], format = "%Y%m%d_%H%M%S")
        f_new <- sprintf("%s.%s", f, when)
        file.rename(f, f_new)
        f_new_info <- file_info(normalizePath(f_new), type = "binary")
        if (debug) logf("- Skipping %s by renaming it to %s", f, f_new_info)
        has_RData <- is_file(f)
        if (!has_RData) {
          warning(sprintf("Skipped %s by renaming it to %s [R_STARTUP_RDATA/startup.rdata=%s]", squote(f_norm), f_new_info, rdata0), call. = FALSE)
        }
      } else if (rdata != "default") {
        warning(sprintf("Ignoring unknown value (%s) of %s/%s",
                        squote(rdata0), squote("R_STARTUP_RDATA"),
                        squote("startup.rdata")),
                call. = FALSE)
      }
    }
  }

  if (debug) {
    if (!no_restore_data) {
      if (has_RData) {
        loads_RData <- TRUE
        logf("- %s", file_info(f, type = "binary"))
      }
    }

    logf("- R_HISTFILE: %s", squote(Sys.getenv("R_HISTFILE")))
    no_restore_history <- any(c("--no-restore-history", "--no-restore", "--vanilla") %in% cmd_args)
    if (!no_restore_history && interactive()) {
      if (is_file(f <- Sys.getenv("R_HISTFILE", "./.Rhistory"))) {
        logf("- %s", file_info(f, type = "txt"))
      }
    }

    where <- find(".First", mode = "function")
    if (where > 0) {
      logf("- .First(): in %s (position %d on search())", squote(search()[where]), where)
    } else if (loads_RData) {
      logf("- .First(): no such function on search(), but it might be that one is defined in the ./RData file")
    } else {
      logf("- .First(): no such function on search()")
    }

    pkgs <- unlist(strsplit(pkgs, split = ",", fixed = TRUE))
    to_be_attached <- !is.element(paste("package:", pkgs, sep = ""), search())
    pkgs <- pkgs[to_be_attached]
    logf("- Remaining packages per R_DEFAULT_PACKAGES to be attached by base::.First.sys() (in order): %s",
         paste(squote(pkgs), collapse = ", "))

    logf("- The R just-in-time (JIT) compiler:")
    jit <- Sys.getenv("R_ENABLE_JIT", NA_character_)
    value <- if (is.na(jit)) "<not set>" else squote(jit)
    logf("  - R_ENABLE_JIT: %s", value)
    if (is.na(jit)) {
      logf("  - JIT compiler enabled by default")
      if (!"compiler" %in% loadedNamespaces()) {
        logf("  - The 'compiler' package will be loaded")
      }
    } else {
      value <- suppressWarnings(as.integer(jit))
      if (!is.na(value) && value > 0L) {
        logf("  - JIT compiler enabled at level %d", jit)
        if (!"compiler" %in% loadedNamespaces()) {
          logf("  - The 'compiler' package will be loaded")
        }
      } else {
        logf("  - JIT compiler disabled")
      }
    }


    logf("The following will be processed when R terminates:")
    for (what in c(".Last", ".Last.sys")) {
      where <- find(what, mode = "function")
      if (where > 0) {
        logf("- %s(): in %s (position %d on search()); circumvented by quit(runLast = FALSE)", what, squote(search()[where]), where)
      } else {
        logf("- %s(): no such function on search()", what)
      }
    }
  }

  invisible(res)
}

#' Checks if running R in RStudio Console or RStudio Terminal
#'
#' @return A logical
#'
#' @references
#' 1. Kevin Ushey (RStudio), Check if R is running in RStudio,
#'    Stackoverflow, 2016-09-23,
#'    \url{https://stackoverflow.com/questions/12389158/check-if-r-is-running-in-rstudio#comment66636507_35849779}
#' 2. Jonathan McPherson (RStudio), Programmatically detect RStudio Terminal vs RStudio Console?,
#'    RStudio Community - RStudio IDE, 2018-01-10,
#'    \url{https://forum.posit.co/t/programmatically-detect-rstudio-terminal-vs-rstudio-console/4107}
is_rstudio_console <- function() {
  (Sys.getenv("RSTUDIO") == "1") && !nzchar(Sys.getenv("RSTUDIO_TERM"))
}

is_rstudio_terminal <- function() {
  (Sys.getenv("RSTUDIO") == "1") &&  nzchar(Sys.getenv("RSTUDIO_TERM"))
}

#' Checks if running R via Jupyter
#'
#' @return A logical
is_jupyter <- function() {
  args <- commandArgs()
  idx <- match("--args", args)
  if (!is.na(idx)) args <- args[seq_len(idx - 1)]
  any(grepl("IRkernel::main()", args))
}

message("*** find_files() ...")


pn <- startup:::find_rprofile()
print(pn)

pn <- startup:::find_rprofile(all = TRUE)
print(pn)

pn <- startup:::find_renviron()
print(pn)

pn <- startup:::find_renviron(all = TRUE)
print(pn)

pn <- startup:::find_rprofile_d()
print(pn)

pn <- startup:::find_rprofile_d(all = TRUE)
print(pn)

pn <- startup:::find_rprofile_d(sibling = TRUE, all = TRUE)
print(pn)

pn <- startup:::find_renviron_d()
print(pn)

pn <- startup:::find_renviron_d(all = TRUE)
print(pn)

pn <- startup:::find_renviron_d(sibling = TRUE, all = TRUE)
print(pn)

path <- system.file(package = "startup")
paths <- file.path(path, c(".Renviron.d", ".Rprofile.d"))
pn <- startup:::list_d_files(paths = paths)
print(pn)

message("*** find_files() ... DONE")

backup <- function(file, quiet = FALSE) {
  if (quiet) notef <- function(...) NULL

  stop_if_not(file.exists(file))
  size <- file_size(file)
  
  timestamp <- format(Sys.time(), "%Y%m%d-%H%M%S")
  backup_file <- sprintf("%s.bak.%s", file, timestamp)
  ## Was another backup file created just before during the same second?
  if (file.exists(backup_file)) {
    timestamp <- format(Sys.time(), "%Y%m%d-%H%M%OS3")
    backup_file <- sprintf("%s.bak.%s", file, timestamp)
  }
  stop_if_not(!file.exists(backup_file))
  res <- file.copy(file, backup_file, overwrite = FALSE)

  backup_size <- file_size(backup_file)
  notef("Backed up R startup file: %s (%d bytes) -> %s (%d bytes)",
        squote(file), size, squote(backup_file), backup_size)
  stop_if_not(file.exists(backup_file), identical(backup_size, size), res)
  
  backup_file
}

#' Register functions to be evaluated at the beginning or end of the R session
#' 
#' @param fcn A function or an R expression. The function must accept zero
#' or more arguments (currently not used). If an expression, it will
#' automatically we wrapped up in an anonymous function.
#'
#' @param append If TRUE (default), the function will be evaluated after
#' previously registered ones, otherwise prepended.
#'
#' @param replace if TRUE, the function replaces any previously registered
#' ones, otherwise it will be added (default).
#'
#' @return (invisible) the list of registered functions.
#'
#' @details
#' These functions register one or more functions to be called when the
#' current R session begins or ends.  The functions are evaluated in a local
#' environment and without exception handlers, which means that if one
#' produces an error, then none of the succeeding functions will be called.
#'
#' To list currently registered functions, use `fcns <- on_session_enter()`
#' or `fcns <- on_session_exit()`.
#' To remove all registered functions, use `on_session_enter(replace = TRUE)`
#' or `on_session_exit(replace = TRUE)`.
#'
#' The `on_session_enter()` function works by recording all `fcn`:s in an
#' internal list which will be evaluated via a custom
#' \code{\link[base:.First]{.First()}} function created in the global
#' environment. Any other `.First()` function on the search path, including
#' a pre-existing `.First()` function in the global environment, is called
#' at the end after registered functions have been called.
#'
#' The `on_session_exit()` function works by recording all `fcn`:s in an
#' internal list which will be evaluated via a custom function that is called
#' when the global environment is garbage collected, which happens at the very
#' end of the R shutdown process.
#' Contrary to a \code{\link[base:.Last]{.Last()}} function, which is not be
#' called if `quit(runLast = FALSE)` is used, functions registered via
#' `on_session_exit()` are always processed.
#' Registered `on_session_exit()` functions are called _after_ `quit()` saves
#' any workspace image to file (`./.RData`), and _after_ any `.Last()` has
#' been called.
#'
#' @examples
#' \dontrun{
#' ## Summarize interactive session upon termination
#' if (interactive()) {
#'   startup::on_session_exit(local({
#'     t0 <- Sys.time()
#'     function(...) {
#'       dt <- difftime(Sys.time(), t0, units = "auto")
#'       msg <- c(
#'         "Session summary:",
#'         sprintf(" * R version: %s", getRversion()),
#'         sprintf(" * Process ID: %d", Sys.getpid()),
#'         sprintf(" * Wall time: %.2f %s", dt, attr(dt, "units"))
#'       )
#'       msg <- paste(msg, collapse = "\n")
#'       message(msg)
#'     }
#'   }))
#' }
#' }
#'
#' @export
on_session_enter <- local({
  .First <- function() {
    "This function was added by startup::on_session_enter()"
    "Evaluate registered functions, cf. environment(.First)$fcns"
    for (fcn in fcns) {
      local(eval(fcn(), envir = parent.frame()))
    }
    
    ## Call any pre-existing .First() on the search path
    "Call any pre-existing .First() on the search path, including"
    "any pre-existing .First() function, cf. environment(.First)$first"
    
    ## Is there a .First() on the search() path excluding existing one
    ## in the global environment?
    e <- globalenv()
    while (!identical(e <- parent.env(e), emptyenv())) {
      if (exists(".First", mode = "function", envir = e, inherits = FALSE)) {
        first <- get(".First", mode = "function", envir = e, inherits = FALSE)
        break
      }
    }
    
    if (is.function(first)) first()
  } ## .First()
  
  function(fcn = NULL, append = TRUE, replace = FALSE) {
    stopifnot(is.logical(append), length(append) == 1L, !is.na(append))
    stopifnot(is.logical(replace), length(replace) == 1L, !is.na(replace))

    if (!is.function(fcn)) {
      expr <- fcn
      fcn <- function(...) NULL
      body(fcn) <- expr
    }

    env <- environment(.First)
    
    ## Set up local .First()? (only once)
    if (!isTRUE(env[["on_session_enter"]])) {
      env <- new.env(parent = globalenv())
      env[["first"]] <- NULL
      env[["fcns"]] <- list()
      env[["on_session_enter"]] <- TRUE
      environment(.First) <<- env
    }

    ## Make sure to record any pre-existing .First() in the global environment
    genv <- globalenv()
    if (exists(".First", envir = genv, inherits = FALSE)) {
      first <- get(".First", envir = genv, inherits = FALSE)
      e <- environment(first)
      if (!isTRUE(e[["on_session_enter"]])) env[["first"]] <- first
    }

    fcns <- env[["fcns"]]
    if (is.null(fcn)) return(fcns)

    ## Replace?
    if (replace) fcns <- list()
  
    ## Append or prepend?
    if (!is.null(fcn)) {
      fcn <- list(fcn)
      fcns <- if (append) c(fcns, fcn) else c(fcn, fcns)
    }
    
    assign(".First", .First, envir = genv)

    invisible(fcns)
  }
})


on_session_enter <- local({
  .First <- function() {
    "This function was added by startup::on_session_enter()"
    "Evaluate registered functions, cf. environment(.First)$fcns"
    for (fcn in fcns) {
      local(fcn())
    }
    
    ## Call any pre-existing .First() on the search path
    "Call any pre-existing .First() on the search path, including"
    "any pre-existing .First() function, cf. environment(.First)$first"
    
    ## Is there a .First() on the search() path excluding existing one
    ## in the global environment?
    e <- globalenv()
    while (!identical(e <- parent.env(e), emptyenv())) {
      if (exists(".First", mode = "function", envir = e, inherits = FALSE)) {
        first <- get(".First", mode = "function", envir = e, inherits = FALSE)
        break
      }
    }
    
    if (is.function(first)) first()
  } ## .First()
  
  function(fcn = NULL, append = TRUE, replace = FALSE) {
    stopifnot(is.logical(append), length(append) == 1L, !is.na(append))
    stopifnot(is.logical(replace), length(replace) == 1L, !is.na(replace))

    if (!is.function(fcn)) {
      expr <- fcn
      fcn <- function(...) NULL
      body(fcn) <- expr
    }

    genv <- globalenv()
  
    ## Make sure to record any pre-existing .First() in the global environment
    first <- NULL
    fcns <- list()
    if (exists(".First", envir = genv, inherits = FALSE)) {
      first <- get(".First", envir = genv, inherits = FALSE)
      env <- environment(first)
      if (isTRUE(env[["on_session_enter"]])) {
        first <- env[["first"]]
        fcns <- env[["fcns"]]
      }
    }

    ## Replace?
    if (replace) fcns <- list()
  
    ## Append or prepend?
    if (!is.null(fcn)) {
      fcn <- list(fcn)
      fcns <- if (append) c(fcns, fcn) else c(fcn, fcns)
    }

    env <- new.env(parent = genv)
    env[["first"]] <- first
    env[["fcns"]] <- fcns
    env[["on_session_enter"]] <- TRUE
    environment(.First) <<- env
    
    assign(".First", .First, envir = genv)
  
    invisible(fcns)
  }
})

#' Gets the pathname of the currently running startup script
#'
#' @return A character string
#' 
#' @export
current_script <- function() {
  current_script_pathname()
}

current_script_pathname <- local({
  .pathname <- NA_character_
  function(pathname = NULL) {
    if (is.null(pathname)) {
      pathname <- .pathname
    } else {
      .pathname <<- pathname
    }
    pathname
  }
})

#' Locates the .Rprofile and .Renviron files used during the startup of R
#'
#' @describeIn find_rprofile Locates the \file{.Rprofile} file used during
#' \R startup.
#'
#' @export
#' @keywords internal
find_rprofile <- function(all = FALSE) {
  pathnames <- c(Sys.getenv("R_PROFILE_USER"), "./.Rprofile", "~/.Rprofile")
  pathnames <- drop_user_files_during_check(pathnames)
  find_files(pathnames, all = all)
}

#' @describeIn find_rprofile Locates the \file{.Renviron} file used during
#' \R startup.
#'
#' @export
#' @keywords internal
find_renviron <- function(all = FALSE) {
  pathnames <- c(Sys.getenv("R_ENVIRON_USER"), "./.Renviron", "~/.Renviron")
  pathnames <- drop_user_files_during_check(pathnames)
  find_files(pathnames, all = all)
}

#' Locates the .Rprofile.d and .Renviron.d directories used during the
#' startup of R
#'
#' @describeIn find_rprofile_d Locates the \file{.Rprofile.d} directory used
#' during \R startup.
#'
#' @export
#' @keywords internal
find_rprofile_d <- function(sibling = FALSE, all = FALSE) {
  ## Only include .Rprofile.d directories if a sibling .Rprofile file exists?
  if (sibling) {
    pathnames <- find_rprofile(all = all)
  } else {
    ## The default R startup search path
    pathnames <- c(Sys.getenv("R_PROFILE_USER"), "~/.Rprofile", "./.Rprofile")
    pathnames <- drop_user_files_during_check(pathnames)
  }

  pathnames <- pathnames[nzchar(pathnames)]
  paths <- sprintf("%s.d", pathnames)
  paths <- c(paths, file.path(get_user_dir("config"), "Rprofile.d"))
  paths_d <- find_d_dirs(paths, all = all)
  if (length(paths_d) == 0) {
    logf("Found no corresponding startup directory %s.",
         paste(squote(paths), collapse = ", "))
  } else {
    logf("Found %s directory %s.", squote(.packageName), paste(squote(paths_d), collapse = ", "))
  }
  paths_d
}

#' @describeIn find_rprofile_d Locates the \file{.Renviron.d} directory used
#' during \R startup.
#'
#' @export
#' @keywords internal
find_renviron_d <- function(sibling = FALSE, all = FALSE) {
  ## Only include .Renviron.d directories if a sibling .Renviron file exists?
  if (sibling) {
    pathnames <- find_renviron(all = all)
  } else {
    ## The default R startup search path
    pathnames <- c(Sys.getenv("R_ENVIRON_USER"), "~/.Renviron", "./.Renviron")
    pathnames <- drop_user_files_during_check(pathnames)
  }

  pathnames <- pathnames[nzchar(pathnames)]
  paths <- sprintf("%s.d", pathnames)
  paths <- c(paths, file.path(get_user_dir("config"), "Renviron.d"))
  paths_d <- find_d_dirs(paths, all = all)
  if (length(paths_d) == 0) {
    logf("Found no corresponding startup directory %s.",
         paste(squote(paths), collapse = ", "))
  } else {
    logf("Found %s directory %s.", squote(.packageName), paste(squote(paths_d), collapse = ", "))
  }
  paths_d
}

find_files <- function(pathnames, all = FALSE) {
  pathnames <- pathnames[file.exists(pathnames)]
  pathnames <- pathnames[!file.info(pathnames)$isdir]

  if (!all) {
    pathnames <- if (length(pathnames) == 0) character(0L) else pathnames[1]
  }

  pathnames
} ## find_files()

find_d_dirs <- function(paths, all = FALSE) {
  if (length(paths) == 0) return(character(0))

  paths <- paths[file.exists(paths)]
  paths <- paths[file.info(paths)$isdir]

  ## Drop duplicates
  paths <- unique(paths)

  ## Drop duplicates after path normalization. For example, ~/.Rprofile.d/
  ## and ./.Rprofile.d/ may be the same folder. Also, if ~/.Rprofile.d/ is
  ## a symbolic links to ~/.config/R/startup/Rprofile.d/, they are also
  ## considered duplicates.
  dups <- duplicated(normalizePath(paths, mustWork = FALSE))
  paths <- paths[!dups]

  if (!all) {
    paths <- if (length(paths) == 0) character(0L) else paths[1]
  }

  paths
} ## find_d_dirs()


list_d_files <- function(paths, recursive = TRUE, filter = NULL) {
  ol <- Sys.getlocale("LC_COLLATE")
  on.exit(Sys.setlocale("LC_COLLATE", ol))
  Sys.setlocale("LC_COLLATE", "C")

  ## Keep only the ones that exists
  paths <- paths[file.exists(paths)]

  ## Nothing to do?
  if (length(paths) == 0) return(character(0L))

  ## For each directory, locate files of interest
  files <- NULL
  for (path in paths) {
    files <- c(files, dir(path = path, pattern = "[^~]$",
                          recursive = recursive, all.files = TRUE,
                          full.names = TRUE))
  }

  ## Drop files such as '#file.R#'
  files <- files[!grepl("^#.*#$", basename(files))]

  ## Drop stray files created by R
  ignores <- c(".Rhistory", ".RData")
  files <- files[!is.element(basename(files), ignores)]

  ## Drop stray directories and files created by macOS
  ## Source: https://apple.stackexchange.com/questions/14980
  ignores <- c(".DS_Store", ".Spotlight-V100", ".TemporaryItems",
               ".VolumeIcon.icns", ".apDisk", ".fseventsd")
  files <- files[!is.element(basename(files), ignores)]
  files <- grep("[/\\\\](__MACOSX|[.]Trash|[.]Trashes)[/\\\\]", files, value = TRUE,
                fixed = FALSE, invert = TRUE)
  hidden <- grep("._", basename(files), fixed = TRUE, value = FALSE)
  if (length(hidden) > 0) {
    hidden_files <- files[hidden]
    hidden_names <- sub("^[.]_", "", basename(hidden_files))
    hidden_siblings <- file.path(dirname(hidden_files), hidden_names)
    ## Workaround for Windows (because mix of forward and backward slashes)
    hidden_siblings <- normalizePath(hidden_siblings, mustWork = FALSE)
    files_normalized <- normalizePath(files, mustWork = FALSE)
    drop <- is.element(hidden_siblings, files_normalized)
    hidden_files <- hidden_files[drop]
    files <- setdiff(files, hidden_files)
  }

  ## Drop files based on filename endings
  files <- grep("([.]md|[.]txt|~)$", files, value = TRUE, invert = TRUE)

  ## Drop "hidden" private files and "hidden" private directories
  ## (double period)
  files <- grep("(^|/|\\\\)[.][.]", files, value = TRUE, invert = TRUE)

  ## Drop version control files
  files <- grep("([.]git|[.]gitignore|[.]hg|[.]hgignore|[.]p4config|[.]p4ignore|[.]svn)$", files, value = TRUE, invert = TRUE)

  ## Nothing to do?
  if (length(files) == 0) return(character(0))

  ## Keep only existing files
  files <- files[file.exists(files)]
  files <- files[!file.info(files)$isdir]

  ## Nothing to do?
  if (length(files) == 0) return(character(0))

  ## Drop duplicates
  files_normalized <- normalizePath(files, winslash = "/")
  files <- files[!duplicated(files_normalized)]

  ## Apply filter?
  if (is.function(filter)) {
    files <- filter(files)
  }

  files
} ## list_d_files()


running_r_cmd <- local({
  running <- NA
  function() {
    if (is.na(running)) {
      running <<- nzchar(Sys.getenv("R_CMD"))
    }
    running
  }
})

drop_user_files_during_check <- function(pathnames) {
  if (!running_r_cmd()) return(pathnames)
  grep("~", pathnames, value = TRUE, invert = TRUE)
}

## Because of the .onLoad() hack, it could be that one day R is
## updated such that the package vignettes are not build.  If so,
## we try to detect it here.

message("*** Assert that package vignettes exist ...")
message(paste(loadedNamespaces(), sep = "\n"))
message(paste(Sys.getenv(), sep = "\n"))

## WORKAROUND: On AppVeyor CI, vignettes are dropped / not built,
## and when running covr, R_CMD is not used.
if (getRversion() >= "3.0.2" &&
    length(packageDescription("startup")$VignetteBuilder) &&
    Sys.getenv("R_COVR") == "") {
  vigns <- utils::vignette(package = "startup")
  print(vigns)
  str(vigns)
  stopifnot(nrow(vigns$results) > 0)
}

## Test vignette engine registration
ovalue <- Sys.getenv("R_CMD")
Sys.setenv("R_CMD" = "dummy")
startup:::register_vignette_engine_during_build_only("startup")
Sys.setenv("R_CMD" = ovalue)

message("*** Assert that package vignettes exist ... DONE")

message("*** install() / uninstall() ...")

print(startup:::is_installed())

path <- tempdir()


message("- install() ...")

file <- file.path(path, ".Rprofile")
res <- startup:::is_installed(file)
print(res)
stopifnot(!res)

print(startup:::install(file))
res <- startup:::is_installed(file)
print(res)
stopifnot(res)

res <- tryCatch({
  startup:::install(file, backup = FALSE)
}, warning = identity)
print(res)
stopifnot(inherits(res, "warning"))

cat("# Empty\n", file = file)
print(startup:::install(file, backup = FALSE))
res <- startup:::is_installed(file)
print(res)
stopifnot(res)


message("- uninstall() ...")

print(startup:::uninstall(file))

res <- startup:::is_installed(file)
print(res)
stopifnot(!res)

res <- tryCatch({
  startup:::uninstall(file)
}, warning = identity)
print(res)
stopifnot(inherits(res, "warning"))

message("*** install() / uninstall() ... DONE")

#' Install and uninstall support for .Renviron.d and .Rprofile.d startup
#' directories
#'
#' Install and uninstall support for \file{.Renviron.d} and \file{.Rprofile.d}
#' startup directories by appending / removing one line of code to the
#' \file{~/.Rprofile} file.
#'
#' @param file The pathname where to create or update the \file{.Rprofile}
#' file.
#'
#' @param backup If `TRUE`, a timestamped backup copy of the original file is
#' created before modifying / overwriting it, otherwise not.  If the backup
#' fails, then an error is produced and the R startup file is unmodified.
#'
#' @param overwrite If the R startup file already exist, then `FALSE` (default)
#' appends the startup code to the end of the file. is overwritten.  If `TRUE`,
#' any pre-existing R startup file is overwritten.
#'
#' @param path The folder where to create \file{.Renviron.d} and
#' \file{.Rprofile.d} directory.
#'
#' @param make_dirs If `TRUE` (default), directories \file{.Renviron.d/} and
#' \file{.Rprofile.d/} are created in folder `path`, if missing.
#'
#' @param quiet If `FALSE` (default), detailed messages are generated,
#' otherwise not.
#'
#' @return The pathname of the R startup file modified.
#'
#' @describeIn install injects a `tryCatch(startup::startup(), ...)` call to
#' the \file{.Rprofile} file, which is created if missing.
#'
#' @export
install <- function(file = rprofile_user(), backup = TRUE, overwrite = FALSE,
                    path = dirname(file), make_dirs = TRUE, quiet = FALSE) {
  if (quiet) notef <- function(...) NULL

  if (make_dirs) {
    dir <- file.path(path, ".Rprofile.d")
    if (!file.exists(dir)) {
      notef("Creating R profile directory: %s", squote(dir))
      dir.create(dir, recursive = TRUE, showWarnings = FALSE)
    }
  
    dir <- file.path(path, ".Renviron.d")
    if (!file.exists(dir)) {
      notef("Creating R environment directory: %s", squote(dir))
      dir.create(dir, recursive = TRUE, showWarnings = FALSE)
    }
  }

  if (is_installed(file)) {
    msg <- sprintf("startup::startup() already installed: %s", squote(file))
    notef(msg)
    warning(msg)
    return(file)
  }

  file_exists <- file.exists(file)
  if (backup && file_exists) backup(file, quiet = quiet)

  code <- "tryCatch(startup::startup(), error=function(ex) message(\".Rprofile error: \", conditionMessage(ex)))\n"

  ## If the .Rprofile file does not have a newline at the end, which is
  ## a mistake, make sure that the appended startup code is on its own line
  if (file_exists && !eof_ok(file)) code <- paste("\n", code, sep = "")
  
  cat(code, file = file, append = !overwrite)
  if (file_exists) {
    notef("%s 'startup::startup()' to already existing R startup file: %s",
          if (overwrite) "Appended" else "Added", squote(file))
  } else {
    notef("Created new R startup file with 'startup::startup()': %s",
          squote(file))
  }

  file
}


#' @describeIn install Remove calls to `startup::startup()` and similar.
#' @export
uninstall <- function(file = rprofile_user(), backup = TRUE, quiet = FALSE) {
  if (quiet) notef <- function(...) NULL

  if (!is_installed(file)) {
    msg <- sprintf("startup::startup() not installed: %s", squote(file))
    notef(msg)
    warning(msg)
    return(file)
  }

  bfr <- readLines(file, warn = FALSE)
  pattern <- "startup::(startup|renviron|rprofile)[(].*[)]"
  bfr2 <- grep(pattern, bfr, value = TRUE, invert = TRUE)
  ## Nothing to do?
  if (isTRUE(all.equal(bfr2, bfr))) {
    msg <- sprintf("startup::startup() not installed: %s", squote(file))
    notef(msg)
    warning(msg)
    return(file)
  }
  if (backup) backup(file, quiet = quiet)
  writeLines(bfr2, con = file)
  notef("R startup file updated: %s", squote(file))

  file
}


rprofile_user <- function() {
  file <- Sys.getenv("R_PROFILE_USER", NA_character_)
  if (!is.na(file) && nzchar(file)) {
    names(file) <- "R_PROFILE_USER"
    return(file)
  }
  file <- file.path("~", ".Rprofile")
  names(file) <- "~/.Rprofile"
  file
}

is_installed <- function(file = rprofile_user()) {
  if (!file.exists(file)) return(FALSE)
  bfr <- readLines(file, warn = FALSE)
  bfr <- gsub("#.*", "", bfr)
  pattern <- "startup::(startup|renviron|rprofile)[(].*[)]"
  res <- any(grepl(pattern, bfr))
  attr(res, "file") <- file
  res
}

## Checks if running R via Rscript
##
## _WARNING: This function does not work - it always returns FALSE \[1\]._
##
## @return A logical
##
## @details
## This functions use `basename(commandArgs()[1])` to infer whether or not
## \R was launched via \file{Rscript}.  On Windows, both \file{Rscript} and
## \file{Rscript.exe} are recognized and the check is case insensitive.
##
## @references
## 1. <https://github.com/HenrikBengtsson/startup/issues/97>
is_rscript <- local({
  result <- NA
  
  function() {
    if (is.na(result)) {
      cmd_args <- getOption("startup.commandArgs", commandArgs())
      executable <- basename(cmd_args[1])
      if (.Platform$OS.type != "windows") {
        result <<- (executable == "Rscript")
      } else {
        result <<- (tolower(executable) %in% c("rscript", "rscript.exe"))
      }
    }
    result
  }
})

register_vignette_engine_during_build_only <- function(pkgname) {
  # Are vignette engines supported?
  if (getRversion() < "3.0.0") return() # Nope!

  ## HACK: Only register vignette engine 'selfonly' during R CMD build
  if (Sys.getenv("R_CMD") == "") return()

  tools::vignetteEngine("selfonly", package = pkgname, pattern = "[.]md$",
    weave = function(file, ...) {
      output <- sprintf("%s.html", tools::file_path_sans_ext(basename(file)))
      md <- readLines(file)

      title <- grep("%\\VignetteIndexEntry{", md, fixed = TRUE, value = TRUE)
      title <- gsub(".*[{](.*)[}].*", "\\1", title)

      ## Inject vignette title
      md <- c(sprintf("# %s\n\n", title), md)

      html <- commonmark::markdown_html(md,
                                        smart = FALSE,
                                        extensions = "table",
                                        normalize = FALSE)
      
      ## Embed images as <img src="data:image/png;base64...">
      mimes <- list(
        gif = "image/gif",
        jpg = "image/jpeg",
        png = "image/png",
        svg = "image/svg+xml"
      )
      html <- unlist(strsplit(html, split = "\n", fixed = TRUE))
      for (ext in names(mimes)) {
        mime <- mimes[[ext]]
        pattern <- sprintf('(.*[ ]src=")([^"]+[.]%s)(".*)', ext)
        idxs <- grep(pattern, html)
        if (length(idxs) == 0) next
        if (!requireNamespace(stealth <- "base64enc", quietly = TRUE)) {
          stop("This vignette requires the ", sQuote(stealth), 
               " package because it contains a ", sQuote(toupper(ext)), 
               " image")
        }
        ns <- getNamespace(stealth)
        dataURI <- get("dataURI", mode = "function", envir = ns)
        for (idx in idxs) {
          file <- gsub(pattern, "\\2", html[idx])
          uri <- dataURI(file = file, mime = mime)
          html[idx] <- gsub(pattern, sprintf("\\1%s\\3", uri), html[idx])
        }
      }

      ## Inject HTML environment
      html <- c("<!DOCTYPE html>",
                "<html lang=\"en\">",
                "<head>",
                sprintf("<title>%s</title>", title),
                "<style>",
                readLines("incl/clean.css", warn = FALSE),
                "</style>",
                "</head>",
                "<body>", html, "</body>",
                "</html>")

      writeLines(html, con = output)
      output
    },

    tangle = function(file, ...) {
      ## As of R 3.3.2, vignette engines must produce tangled output, but as
      ## long as it contains all comments then 'R CMD build' will drop it.
      output <- sprintf("%s.R", tools::file_path_sans_ext(basename(file)))
      cat(sprintf("### This is an R script tangled from %s\n",
                  sQuote(basename(file))), file = output)
      output
    }
  )
}

filter_files <- startup:::filter_files
sysinfo <- startup:::sysinfo()
sysinfo$os <- "linux"
sysinfo$rstudio <- TRUE
sysinfo$wine <- FALSE
sysinfo$interactive <- FALSE
Sys.setenv(FOO = "abc")
Sys.setenv(BAR = "123")
Sys.unsetenv("UNKNOWN")

message("*** filter_files() ...")

filesets <- list(
  A = c("a" = TRUE, "b" = TRUE, "package=<non-existing-package>" = FALSE,
        "package=startup" = TRUE),
  B = c("a,os=linux" = TRUE, "b,os=windows" = FALSE, "os=linux,os=osx" = FALSE,
        "package=startup,os=linux" = TRUE),
  C = c("a,os=linux" = TRUE, "b,os!=linux" = FALSE, "c,os!=windows" = TRUE),
  D = c("package!=startup" = FALSE, "package!=<non-existing-package>" = TRUE),
  E = c("os=linux,package!=startup" = FALSE,
        "os=linux,package!=<non-existing-package>" = TRUE),
  F = c("a" = TRUE, "/home/alice/.Rprofile.d/package=foo" = FALSE),
  G = c("a" = TRUE, "/home/alice/.Rprofile.d/package!=foo" = TRUE),
  H = c("/home/alice/.Rprofile.d/rstudio=T/wine=0" = TRUE),
  I = c("/home/alice/.Rprofile.d/rstudio!=f/wine!=1" = TRUE),
  J = c("/home/alice/.Rprofile.d/interactive=TRUE" = FALSE,
        "/home/alice/.Rprofile.d/interactive=FALSE" = TRUE),
  K = c("/home/alice/.Rprofile.d/interactive=TRUE/package=fortunes" = FALSE),
  L = c("/home/alice/.Rprofile.d/package=startup,package=base" = TRUE),
  M = c("/home/alice/.Rprofile.d/package=startup,package!=base" = FALSE),
  N = c("/home/alice/.Rprofile.d/FOO=abc,BAR=123" = TRUE),
  O = c("/home/alice/.Rprofile.d/FOO=abc,BAR!=321" = TRUE),
  P = c("/home/alice/.Rprofile.d/FOO=abc,BAR!=123" = FALSE),
  Q = c("/home/alice/.Rprofile.d/FOO=abc/BAR=123/help" = TRUE),
  R = c("/home/alice/.Rprofile.d/UNKNOWN=42/test" = FALSE),
  S = c("/home/alice/.Rprofile.d/UNKNOWN!=42/test" = TRUE)
)

## Test with filename extensions *.R as well
filesets2 <- lapply(filesets, FUN = function(f) {
  names(f) <- sprintf("%s.R", names(f))
  f
})
names(filesets2) <- sprintf("%s.R", names(filesets))
#filesets <- c(filesets, filesets2)


for (kk in seq_along(filesets)) {
  message(sprintf("File set #%d (%s) ...", kk, names(filesets)[kk]))

  files <- filesets[[kk]]
  files_truth <- names(files)[files]
  files <- names(files)
  cat("Before:\n")
  print(files)

  files_filtered <- filter_files(files, info = sysinfo)
  cat("After:\n")
  print(files_filtered)
  cat("Expected:\n")
  print(files_truth)
  
  stopifnot(all.equal(files_filtered, files_truth, check.attributes = FALSE))

  message(sprintf("File set #%d (%s) ... DONE", kk, names(filesets)[kk]))
}

message("*** filter_files() ... DONE")

#' Check if running R via radian (formerly known as rtichoke and rice)
#'
#' @return A logical
#'
#' @references
#' 1. radian - A 21 century R console (previously known as rtichoke and rice),
#'    \url{https://github.com/randy3k/radian}
is_radian <- function() {
  ## radian (>= 0.3.0)
  if (nzchar(Sys.getenv("RADIAN_VERSION"))) return(TRUE)
  
  ## rtichoke (< 0.3.0)
  if (nzchar(Sys.getenv("RTICHOKE_VERSION"))) return(TRUE)
  
  ## rice (<= 0.1.1)
  if (nzchar(Sys.getenv("RICE_VERSION"))) return(TRUE)
  ## rice (< 0.0.9)
  tolower(basename(Sys.getenv("_"))) == "rice"
}

#' Restarts R
#'
#' Restarts \R by quitting the current \R session and launching a new one.
#'
#' @param status An integer specifying the exit code of the current
#' \R session.
#' 
#' @param workdir The working directory where the new \R session should
#' be launched from.  If `NULL`, then the working directory that was in
#' place when the \pkg{startup} package was first loaded.  If using
#' `startup::startup()` in an \file{.Rprofile} startup file, then this
#' is likely to record the directory from which \R itself was launched from.
#' 
#' @param rcmd A character string specifying the command for launching \R.
#' The default is the same as used to launch the current \R session, i.e.
#' \code{\link[base:commandArgs]{commandArgs()[1]}}.
#'
#' @param args A character vector specifying zero or more command-line
#' arguments to be appended to the system call of \code{rcmd}.
#'
#' @param envvars A named character vector of environment variables to
#' be set when calling \R.
#'
#' @param as A character string specifying a predefined setups of `rcmd`,
#' `args`, and `envvars`.  For details, see below.
#' 
#' @param quiet Should the restart be quiet or not?
#' If `NA` and `as == "current"`, then `quiet` is `TRUE` if the current
#' \R session was started quietly, otherwise `FALSE`.
#' 
#' @param debug If `TRUE`, debug messages are outputted, otherwise not.
#'
#' @section Predefined setups:
#' Argument `as` may take the following values:
#' \describe{
#'  \item{\code{"current"}:}{(Default) A setup that emulates the setup of the
#'   current \R session as far as possible by relaunching \R with the same
#'   command-line call (= [base::commandArgs()]).
#'  }
#'  \item{\code{"specified"}:}{According to `rcmd`, `args`, and `envvars`.}
#'  \item{\code{"R CMD build"}:}{A setup that emulates
#'   [`R CMD build`](https://github.com/wch/r-source/blob/R-3-4-branch/src/scripts/build)
#'   as far as possible.
#'  }
#'  \item{\code{"R CMD check"}:}{A setup that emulates
#'   [`R CMD check`](https://github.com/wch/r-source/blob/R-3-4-branch/src/scripts/check)
#'   as far as possible, which happens to be identical to the
#'  `"R CMD build"` setup.
#'  }
#'  \item{\code{"R CMD INSTALL"}:}{A setup that emulates
#'   [`R CMD INSTALL`](https://github.com/wch/r-source/blob/R-3-4-branch/src/scripts/INSTALL)
#'   as far as possible.
#'  }
#' }
#' If specified, command-line arguments in `args` and environment variables
#' in `envvars` are _appended_ accordingly.
#'
#' @section Known limitations:
#' It is _not_ possible to restart an \R session running in _RGui_ on
#' Microsoft Windows using this function.
#' 
#' It is _not_ possible to restart an \R session in the RStudio _Console_
#' using this function.  However, it does work when running \R in the
#' RStudio _Terminal_.
#' 
#' RStudio provides `rstudioapi::restartSession()` which will indeed restart
#' the RStudio Console.  However, it does not let you control how \R is
#' restarted, e.g. with what command-line options and what environment
#' variables.  Importantly, the new \R session will have the same set of
#' packages loaded as before, the same variables in the global environment,
#' and so on.
#'
#' @examples
#' \dontrun{
#'   ## Relaunch R with debugging of startup::startup() enabled
#'   startup::restart(envvars = c(R_STARTUP_DEBUG = TRUE))
#'
#'   ## Relaunch R without loading user Rprofile files
#'   startup::restart(args = "--no-init-file")
#'
#'   ## Mimic 'R CMD build' and 'R CMD check'
#'   startup::restart(as = "R CMD build")
#'   startup::restart(as = "R CMD check")
#'   ## ... which are both short for
#'   startup::restart(args = c("--no-restore"),
#'                    envvars = c(R_DEFAULT_PACKAGES="", LC_COLLATE="C"))
#' }
#'
#' @export
restart <- function(status = 0L,
                    workdir = NULL,
                    rcmd = NULL, args = NULL, envvars = NULL,
                    as = c("current", "specified",
                           "R CMD build", "R CMD check", "R CMD INSTALL"),
                    quiet = FALSE,
                    debug = NA) {
  debug(debug)
  logf("Restarting R ...")

  ## The RStudio Console cannot be restarted this way
  if (is_rstudio_console()) {
    stop("R sessions run via the RStudio Console cannot be restarted using startup::restart(). It is possible to restart R in an RStudio Terminal. To restart an R session in the RStudio Console, use rstudioapi::restartSession().")
  }

  ## The Windows RGui cannot be restarted this way
  if (sysinfo()$gui == "Rgui") {
    stop("R sessions run via the Windows RGui cannot be restarted using startup::restart().")
  }
  
  if (is.null(workdir)) {
    workdir <- startup_session_options()$startup.session.startdir
  }
  if (!is_dir(workdir)) {
    stop("Argument 'workdir' specifies a non-existing directory: ",
         squote(workdir))
  }
  
  cmdargs <- commandArgs()

  if (is.null(rcmd)) rcmd <- cmdargs[1]
  stop_if_not(length(rcmd) == 1L, is.character(rcmd))
  rcmd_t <- Sys.which(rcmd)
  if (rcmd_t == "") {
    stop("Argument 'rcmd' specifies a non-existing command: ", squote(rcmd))
  }

  as <- match.arg(as)
  if (as == "specified") {
  } else if (as == "current") {
    if (is.null(args)) {
      ## WORKAROUND: When running 'radian', commandArgs() does not
      ## reflect how it was started.
      ## https://github.com/randy3k/radian/issues/23#issuecomment-375078246
      if (is_radian()) {  
        args <- Sys.getenv("RADIAN_COMMAND_ARGS")
      } else {
        args <- cmdargs[-1]
      }
    }
    ## Restart quietly if current session was start quietly?
    if (is.na(quiet)) quiet <- any(args %in% c("--quiet", "-q"))
  } else if (as %in% c("R CMD build", "R CMD check")) {
    ## Source:
    ##  - src/scripts/build
    ##  - src/scripts/check
    ## Also '--slave', but we disable that for now to make it clear
    ## that the session is restarted.

    if (is_radian()) {
      stop(sprintf("startup::restart(as = %s) is not supported when running R via radian", dQuote(as)))
    }
    
    args <- c("--no-restore", args)
    envvars <- c(R_DEFAULT_PACKAGES = "", LC_COLLATE = "C", envvars)
  } else if (as %in% c("R CMD INSTALL")) {
    ## Source:
    ##  - src/scripts/INSTALL
    ## Also '--slave', but we disable that for now to make it clear
    ## that the session is restarted.

    if (is_radian()) {
      stop(sprintf("startup::restart(as = %s) is not supported when running R via radian", dQuote(as)))
    }
    
    vanilla_install <- nzchar(Sys.getenv("R_INSTALL_VANILLA"))
    if (vanilla_install) {
      args <- c("--vanilla", args)
    } else {
      args <- c("--no-restore", args)
    }
    envvars <- c(R_DEFAULT_PACKAGES = "", LC_COLLATE = "C", envvars)
  } else {
    stop("Unknown value on argument 'as': ", squote(as))
  }

  ## Restart quietly or not?
  if (as != "specified") {
    if (quiet) {
      if (is_radian()) {
        ## Only radian (>= 0.2.8) supports '--quiet'
        ## Comment: prior to v0.3.0, radian was actually called 'rtichoke',
        ## so let's just require 'radian' here.
        version <- Sys.getenv("RADIAN_VERSION")
        stopifnot(nzchar(version))
        version <- package_version(version)
        if (version < "0.2.8") {
          stop("startup::restart(quiet = TRUE) requires radian (>= 0.2.8)")
        }
      }
      args <- c("--quiet", args)
    } else {
      args <- setdiff(args, "--quiet")
    }
  }
  
  if (!is.null(envvars) && length(envvars) > 0L) {
    stop_if_not(!is.null(names(envvars)))
    envvars <- sprintf("%s=%s", names(envvars), shQuote(envvars))
  }

  ## To please R CMD check
  envir <- globalenv()

  ## Make sure to call existing .Last(), iff any
  has_last <- exists(".Last", envir = envir, inherits = FALSE)
  if (has_last) {
    last_org <- get(".Last", envir = envir, inherits = FALSE)
  } else {
    last_org <- function() NULL
  }

  logf("- R executable: %s", rcmd)
  logf("- Command-line arguments: %s", paste(args, collapse = " "))
  logf("- Environment variables: %s", paste(envvars, collapse = " "))
  
  assign(".Last", function() {
    last_org()
    system2(rcmd, args = args, env = envvars)
  }, envir = envir)

  logf("- quitting current R session")
  if (has_last) logf("- existing .Last() will be acknowledged")
  logf("Restarting R ... done")

  setwd(workdir)
  quit(save = "no", status = status, runLast = TRUE)
}

message("*** sysinfo() ...")

message("- Base R system information:")

utils::str(list(
  "capabilities()"     = as.list(capabilities()),
  .Machine             = .Machine,
  .Platform            = .Platform,
  R.version            = R.version,
  "Sys.info()"         = as.list(Sys.info())
))

message("- Session information:")
print(utils::sessionInfo())


message("- sysinfo():")
print(startup::sysinfo())

message("*** sysinfo() ... DONE")

get_os <- function() {
  if (.Platform[["OS.type"]] == "windows") return("windows")
  if (Sys.info()["sysname"] == "Darwin") return("macos")
  if (.Platform[["OS.type"]] == "unix") return("unix")
  NA_character_
}


## REFERENCES:
## [1] https://docs.microsoft.com/en-us/windows/deployment/usmt/usmt-recognized-environment-variables
get_windows_local_appdata <- function() {
  ## Env var 'LOCALAPPDATA' should be defined on Windows Vista and later
  path <- Sys.getenv("LOCALAPPDATA", NA_character_)
  if (!is.na(path)) return(path)

  ## We might end up here because we're on Windows XP or for other reasons
  root <- Sys.getenv("USERPROFILE", NA_character_)
  if (is.na(root)) return(NA_character_)  ## Should not happen

  ## Is there an 'AppData/Local' folder?
  path <- file.path(root, "AppData" , "Local")
  if (!is.na(path)) return(path)

  ## Is there an 'Local Settings/Application Data' folder? (Windows XP)
  path <- file.path(root, "Local Settings", "Application Data")
  if (!is.na(path)) return(path)

  ## We shouldn't really end up here, but who knows ...
  NA_character_
}


find_cache_path <- function(dirs = c("R", utils::packageName())) {
  ## Per tools::R_user_dir(which = "cache") of R (>= 4.0.0)
  root <- Sys.getenv("R_USER_CACHE_DIR", NA_character_)
  if (is.na(root)) {
    root <- Sys.getenv("XDG_CACHE_HOME", NA_character_)
  }
  if (is.na(root)) {
    os <- get_os()
    root <- switch(os,
      windows = file.path(get_windows_local_appdata(), "R", "cache"),
      macos = file.path("~", "Library", "Caches", "org.R-project.R"),
      unix = file.path("~", ".cache"),
      NA_character_
    )
  }

  ## Failed to find a OS-specific cache folder?
  if (is.na(root)) {
    os <- get_os()
    stop("Failed to locate local cache folder on this operating system: ", os)
  }
  
  path <- c(root, dirs)
  path <- do.call(file.path, args = as.list(path))
  path <- normalizePath(path, mustWork = FALSE)

  path
}


find_config_path <- function(dirs = c("R", utils::packageName())) {
  ## Per tools::R_user_dir(which = "config") of R (>= 4.0.0)
  root <- Sys.getenv("R_USER_CONFIG_DIR", NA_character_)
  if (is.na(root)) {
    root <- Sys.getenv("XDG_CONFIG_HOME", NA_character_)
  }
  if (is.na(root)) {
    os <- get_os()
    root <- switch(os,
      windows = file.path(get_windows_local_appdata(), "R", "config"),
      macos = file.path("~", "Library", "Preferences", "org.R-project.R"),
      unix = file.path("~", ".config"),
      NA_character_
    )
  }

  ## Failed to find a OS-specific cache folder?
  if (is.na(root)) {
    os <- get_os()
    stop("Failed to locate local cache folder on this operating system: ", os)
  }
  
  path <- c(root, dirs)
  path <- do.call(file.path, args = as.list(path))
  path <- normalizePath(path, mustWork = FALSE)

  path
}

get_user_dir <- local({
  paths <- list()
  function(which = c("cache", "config"), create = FALSE) {
    which <- match.arg(which)
    path <- paths[[which]]
    if (is.null(path)) {
      path <- switch(which,
        cache = find_cache_path(),
        config = find_config_path()
      )
      paths[[which]] <<- path
    }

    if (create && !is_dir(path)) {
      dir.create(path, recursive = TRUE)
      stop_if_not(is_dir(path))
    }

    path
  }
})

renviron_d <- startup::renviron_d
rprofile_d <- startup::rprofile_d

message("*** startup() ...")

message("*** renviron_d() ...")

paths <- system.file("Renviron.d", package = "startup")
print(paths)
api <- renviron_d(paths = paths, skip = FALSE)
str(api)
renviron_d(paths = paths, skip = FALSE, dryrun = TRUE)

message("*** renviron_d() ... DONE")

message("*** rprofile_d() ...")

paths <- system.file("Rprofile.d", package = "startup")
print(paths)
api <- rprofile_d(paths = paths, skip = FALSE)
str(api)
rprofile_d(paths = paths, skip = FALSE, dryrun = TRUE)

message("*** rprofile_d() ... DONE")


message("*** startup() - exceptions ...")

path <- system.file("Rprofile.d,checks", package = "startup")
path_tmp <- tempdir()
file.copy(path, path_tmp, recursive = TRUE, overwrite = TRUE)

oopts <- options(encoding = "native.enc")
res <- tryCatch({
  rprofile_d(paths = path_tmp, skip = FALSE, on_error = "warning")
}, warning = identity)
stopifnot(inherits(res, "simpleWarning"))

options(encoding = "native.enc")
res <- tryCatch({
  rprofile_d(paths = path_tmp, skip = FALSE, on_error = "immediate.warning")
}, warning = identity)
stopifnot(inherits(res, "simpleWarning"))

options(encoding = "native.enc")
res <- tryCatch({
  rprofile_d(paths = path_tmp, skip = FALSE, on_error = "error")
}, error = identity)
stopifnot(inherits(res, "simpleError"))

options(encoding = "native.enc")
res <- rprofile_d(paths = path_tmp, skip = FALSE, on_error = "message")
stopifnot(is.list(res))

options(encoding = "native.enc")
res <- rprofile_d(paths = path_tmp, skip = FALSE, on_error = "ignore")
stopifnot(is.list(res))

options(oopts)

message("*** startup() - exceptions ... DONE")


message("*** startup() ... DONE")

message("*** Checks ...")

startup::check(fix = FALSE)
startup::check(all = TRUE, fix = FALSE)

startup:::check_rprofile_eof()
startup:::check_rprofile_eof(all = TRUE)

startup:::check_rprofile_update_packages()
startup:::check_rprofile_update_packages(all = TRUE)

startup:::check_options()

oopts <- options(encoding = "C")
res <- tryCatch({
  startup:::check_options()
}, warning = identity)
if (!interactive()) stopifnot(inherits(res, "simpleWarning"))
options(oopts)

message("*** Checks - test files ...")

path <- system.file("Rprofile.d,checks", package = "startup")
path_tmp <- tempdir()
file.copy(path, path_tmp, recursive = TRUE, overwrite = TRUE)

files <- startup:::list_d_files(path_tmp)
print(files)

res <- tryCatch({
  startup:::check_rprofile_eof(files = files, fix = FALSE)
}, error = identity)
stopifnot(inherits(res, "simpleError"))

res <- tryCatch({
  startup:::check_rprofile_eof(files = files)
}, warning = identity)
stopifnot(inherits(res, "simpleWarning"))

res <- tryCatch({
  startup:::check_rprofile_update_packages(files = files)
}, error = identity)
stopifnot(inherits(res, "simpleError"))


message("*** Checks - test files ... DONE")

message("*** File name capitalization ...")

for (what in c("Renviron", "Rprofile")) {
  file <- sprintf(".%s", what)
  pathname <- file.path(tempdir(), file)
  cat("dummy", file = pathname)
  res <- startup:::warn_file_capitalization(pathname, what)
  stopifnot(isTRUE(res))
  file.remove(pathname)
  
  file <- toupper(file)
  pathname <- file.path(tempdir(), file)
  cat("dummy", file = pathname)
  res <- startup:::warn_file_capitalization(pathname, what)
  stopifnot(!isTRUE(res))

  res <- tryCatch({
    startup:::warn_file_capitalization(pathname, what)
  }, warning = identity)
  stopifnot(inherits(res, "warning"))
  file.remove(pathname)
}

message("*** File name capitalization ... DONE")

message("*** Checks ... DONE")

message("*** debug() ...")

debug <- startup:::debug ## non-exported
is_debug_on <- startup::is_debug_on

print(debug())
print(is_debug_on())

print(debug(TRUE))
stopifnot(is_debug_on())

startup:::log("Hello")
startup:::logf("Hello %s", "world")
startup:::logp(str(1:10))

print(debug(FALSE))
stopifnot(!is_debug_on())

message("*** debug() ... DONE")

rm(list = "debug")

options(startup.dryrun = TRUE)

message("*** api() ...")

message("*** api()")
api <- startup:::api()
stopifnot(is.list(api), length(api) > 0)

message("*** renviron_d()")
res <- startup::renviron_d()
stopifnot(all.equal(res, api))

message("*** rprofile_d()")
res <- startup::rprofile_d()
stopifnot(all.equal(res, api))

res <- startup::rprofile_d(all = TRUE)

message("*** startup(unload = FALSE)")
res <- startup::startup(unload = FALSE)
stopifnot(all.equal(res, api))

message("*** startup(debug = TRUE)")
## Fake some settings / calls to cover more code
options(startup.debug.commandArgs = commandArgs()[1])
save.image(file = ".RData")
res <- startup::startup(debug = TRUE)
str(res)
stopifnot(all.equal(res, api))
file.remove(".RData")
options(startup.debug.commandArgs = NULL)

options(startup.dryrun = FALSE)

message("*** api() ... DONE")

rm(list = c("api", "res"))

#' Record R session information as options
#'
#' @param action If `"update"` or `"overwrite"`, \R options
#' `"startup.session.*"` are set.  If `"update"`, then such options that are
#' not already set are updated.  If `"erase"`, any existing
#' `"startup.session.*"` options are removed.
#'
#' @return Returns invisibly a named list of the options prefixed
#' `"startup.session."`:
#' \describe{
#'   \item{`startup.session.startdir`}{(character) the working directory when
#'     the \pkg{startup} was first loaded.  If `startup::startup()` is called
#'     at the very beginning of the \file{.Rprofile} file, this is also the
#'     directory that the current \R session was launched from.}
#'   \item{`startup.session.starttime`}{(POSIXct) the time when the
#'     \pkg{startup} was first loaded.}
#'   \item{`startup.session.id`}{(character) a unique ID for the current \R
#'     session.}
#'   \item{`startup.session.dumpto`}{(character) a session-specific name that
#'     can be used for argument `dumpto` of [dump.frames()][utils::dump.frames]
#'     (also for dumping to file).}
#' }
#'
#' @examples
#' opts <- startup::startup_session_options()
#' opts
#'
#' @export
startup_session_options <- function(action = c("update", "overwrite", "erase")) {
  action <- match.arg(action)

  if (action %in% c("update", "overwrite")) {
    pwd <- getwd()
    id <- basename(tempdir())
    time <- Sys.time()

    starttime_iso <- format(time, format = "%Y%m%d-%H%M%S")
    opts <- list(
      startup.session.startdir = pwd,
      startup.session.starttime = time,
      startup.session.starttime_iso = starttime_iso,
      startup.session.id = id,
      startup.session.dumpto = file.path(pwd, sprintf("last.dump_%s_%s",
                                                      starttime_iso, id))
    )

    ## Reuse existing startup.session.* options?
    if (action == "update") {
      old_opts <- options()
      old_opts <- old_opts[grep("^startup[.]session[.]", names(old_opts))]
      for (name in names(old_opts)) opts[[name]] <- old_opts[[name]]
    }

    do.call(options, args = opts)
  } else if (action == "erase") {
    opts <- options()
    opts <- opts[grep("^startup[.]session[.]", names(opts))]
    for (name in names(opts)) opts[name] <- NULL
    do.call(options, args = opts)
  }

  invisible(opts)
}

message("*** startup_session_options() ...")

opts_0 <- startup::startup_session_options()
utils::str(opts_0)

opts_1 <- startup::startup_session_options(action = "update")
utils::str(opts_1)
stopifnot(identical(opts_1, opts_0))

opts_2 <- startup::startup_session_options(action = "overwrite")
utils::str(opts_2)

opts_3 <- startup::startup_session_options(action = "erase")
utils::str(opts_3)
stopifnot(all(unlist(lapply(opts_3, FUN = is.null))))

opts_4 <- startup::startup_session_options()
utils::str(opts_4)

message("*** startup_session_options() ... DONE")

# CRAN submission startup 0.22.0

on 2024-07-29


## Notes not sent to CRAN

### R CMD check validation

The package has been verified using `R CMD check --as-cran` on:

| R version | GitHub | mac/win-builder |
| --------- | ------ | --------------- |
| 3.5.x     | L      |                 |
| 4.2.x     | L      |                 |
| 4.3.x     | L M  W |                 |
| 4.4.x     | L M  W | M1 W            |
| devel     | L M1 W |    W            |

_Legend: OS: L = Linux, M = macOS, M1 = macOS M1, W = Windows_

# Contributing to the 'startup' package

This Git repository uses the [Git Flow](https://nvie.com/posts/a-successful-git-branching-model/) branching model (the [`git flow`](https://github.com/petervanderdoes/gitflow-avh) extension is useful for this).  The [`develop`](https://github.com/HenrikBengtsson/startup/tree/develop) branch contains the latest contributions and other code that will appear in the next release, and the [`master`](https://github.com/HenrikBengtsson/startup) branch contains the code of the latest release, which is exactly what is currently on [CRAN](https://cran.r-project.org/package=startup).

Contributing to this package is easy.  Just send a [pull request](https://help.github.com/articles/using-pull-requests/).  When you send your PR, make sure `develop` is the destination branch on the [startup repository](https://github.com/HenrikBengtsson/startup).  Your PR should pass `R CMD check --as-cran`, which will also be checked by  <a href="https://github.com/HenrikBengtsson/startup/actions?query=workflow%3AR-CMD-check">GitHub Actions</a> and  when the PR is submitted.

We abide to the [Code of Conduct](https://www.contributor-covenant.org/version/2/0/code_of_conduct/) of Contributor Covenant.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check.R
\name{check}
\alias{check}
\title{Check for and fix common mistakes in .Rprofile}
\usage{
check(all = FALSE, fix = TRUE, backup = TRUE, debug = FALSE)
}
\arguments{
\item{all}{Should all or only the first entry on
\link[base:Startup]{the R startup search path} be checked?}

\item{fix}{If \code{TRUE}, detected issues will be tried to be automatically
fixed, otherwise not.}

\item{backup}{If \code{TRUE}, a timestamped backup copy of the original file is
created before modifying it, otherwise not.}

\item{debug}{If \code{TRUE}, debug messages are outputted, otherwise not.}
}
\value{
Returns invisibly a character vector of files that were "fixed"
(modified), if any.  If no files needed to be fixed, or \code{fix = TRUE},
then an empty vector is returned.
}
\description{
Check for and fix common mistakes in \file{.Rprofile} files.
}
\references{
\enumerate{
\item R-devel thread 'Last line in .Rprofile must have newline (PR#4056)',
2003-09-03,
\url{https://stat.ethz.ch/pipermail/r-devel/2003-September/027457.html}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/agenda.R
\name{reset_when_cache}
\alias{reset_when_cache}
\title{Reset all or parts of the "when" cache}
\usage{
reset_when_cache(
  when = c("once", "hourly", "daily", "weekly", "fortnightly", "monthly")
)
}
\arguments{
\item{when}{(character vector) Specifies for which "when" frequencies
the cache should be reset.  The default is to reset all of them.}
}
\value{
(invisible) The pathnames of the "when" cache files removed.
}
\description{
Reset all or parts of the "when" cache
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_webr.R
\name{is_webr}
\alias{is_webr}
\title{Checks if running R via webR}
\usage{
is_webr()
}
\value{
A logical
}
\description{
Checks if running R via webR
}
\references{
\enumerate{
\item WebR - R in the Browser, \url{https://docs.r-wasm.org/webr/latest/}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_jupyter.R
\name{is_jupyter}
\alias{is_jupyter}
\title{Checks if running R via Jupyter}
\usage{
is_jupyter()
}
\value{
A logical
}
\description{
Checks if running R via Jupyter
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/current_script.R
\name{current_script}
\alias{current_script}
\title{Gets the pathname of the currently running startup script}
\usage{
current_script()
}
\value{
A character string
}
\description{
Gets the pathname of the currently running startup script
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_vscode.R
\name{is_vscode}
\alias{is_vscode}
\title{Checks if running R via Visual Studio Code (VSCode)}
\usage{
is_vscode()
}
\value{
A logical
}
\description{
Checks if running R via Visual Studio Code (VSCode)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_microsoftr.R
\name{is_microsoftr}
\alias{is_microsoftr}
\title{Check if running R via Microsoft R Open}
\usage{
is_microsoftr()
}
\value{
A logical
}
\description{
Check if running R via Microsoft R Open
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_pqr.R
\name{is_pqr}
\alias{is_pqr}
\title{Check if running a Pretty Quick Version of R (pqR)}
\usage{
is_pqr()
}
\value{
A logical
}
\description{
Check if running a Pretty Quick Version of R (pqR)
}
\references{
\enumerate{
\item pqR - a pretty quick version of R,
\url{http://www.pqr-project.org/}
\item GitHub repository for 'pqR'
\url{https://github.com/radfordneal/pqR}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_files.R
\name{find_rprofile_d}
\alias{find_rprofile_d}
\alias{find_renviron_d}
\title{Locates the .Rprofile.d and .Renviron.d directories used during the
startup of R}
\usage{
find_rprofile_d(sibling = FALSE, all = FALSE)

find_renviron_d(sibling = FALSE, all = FALSE)
}
\description{
Locates the .Rprofile.d and .Renviron.d directories used during the
startup of R
}
\section{Functions}{
\itemize{
\item \code{find_rprofile_d()}: Locates the \file{.Rprofile.d} directory used
during \R startup.

\item \code{find_renviron_d()}: Locates the \file{.Renviron.d} directory used
during \R startup.

}}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/debug.R
\name{is_debug_on}
\alias{is_debug_on}
\title{Checks whether startup debug is on or not}
\usage{
is_debug_on()
}
\value{
Returns \code{TRUE} is debug is enabled and \code{FALSE} otherwise.
}
\description{
Checks whether startup debug is on or not
}
\details{
The debug mode is when \code{\link[=startup]{startup()}} is called, either explicitly
via argument \code{debug} or via environment variable \code{R_STARTUP_DEBUG}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/restart.R
\name{restart}
\alias{restart}
\title{Restarts R}
\usage{
restart(
  status = 0L,
  workdir = NULL,
  rcmd = NULL,
  args = NULL,
  envvars = NULL,
  as = c("current", "specified", "R CMD build", "R CMD check", "R CMD INSTALL"),
  quiet = FALSE,
  debug = NA
)
}
\arguments{
\item{status}{An integer specifying the exit code of the current
\R session.}

\item{workdir}{The working directory where the new \R session should
be launched from.  If \code{NULL}, then the working directory that was in
place when the \pkg{startup} package was first loaded.  If using
\code{startup::startup()} in an \file{.Rprofile} startup file, then this
is likely to record the directory from which \R itself was launched from.}

\item{rcmd}{A character string specifying the command for launching \R.
The default is the same as used to launch the current \R session, i.e.
\code{\link[base:commandArgs]{commandArgs()[1]}}.}

\item{args}{A character vector specifying zero or more command-line
arguments to be appended to the system call of \code{rcmd}.}

\item{envvars}{A named character vector of environment variables to
be set when calling \R.}

\item{as}{A character string specifying a predefined setups of \code{rcmd},
\code{args}, and \code{envvars}.  For details, see below.}

\item{quiet}{Should the restart be quiet or not?
If \code{NA} and \code{as == "current"}, then \code{quiet} is \code{TRUE} if the current
\R session was started quietly, otherwise \code{FALSE}.}

\item{debug}{If \code{TRUE}, debug messages are outputted, otherwise not.}
}
\description{
Restarts \R by quitting the current \R session and launching a new one.
}
\section{Predefined setups}{

Argument \code{as} may take the following values:
\describe{
\item{\code{"current"}:}{(Default) A setup that emulates the setup of the
current \R session as far as possible by relaunching \R with the same
command-line call (= \code{\link[base:commandArgs]{base::commandArgs()}}).
}
\item{\code{"specified"}:}{According to \code{rcmd}, \code{args}, and \code{envvars}.}
\item{\code{"R CMD build"}:}{A setup that emulates
\href{https://github.com/wch/r-source/blob/R-3-4-branch/src/scripts/build}{\verb{R CMD build}}
as far as possible.
}
\item{\code{"R CMD check"}:}{A setup that emulates
\href{https://github.com/wch/r-source/blob/R-3-4-branch/src/scripts/check}{\verb{R CMD check}}
as far as possible, which happens to be identical to the
\code{"R CMD build"} setup.
}
\item{\code{"R CMD INSTALL"}:}{A setup that emulates
\href{https://github.com/wch/r-source/blob/R-3-4-branch/src/scripts/INSTALL}{\verb{R CMD INSTALL}}
as far as possible.
}
}
If specified, command-line arguments in \code{args} and environment variables
in \code{envvars} are \emph{appended} accordingly.
}

\section{Known limitations}{

It is \emph{not} possible to restart an \R session running in \emph{RGui} on
Microsoft Windows using this function.

It is \emph{not} possible to restart an \R session in the RStudio \emph{Console}
using this function.  However, it does work when running \R in the
RStudio \emph{Terminal}.

RStudio provides \code{rstudioapi::restartSession()} which will indeed restart
the RStudio Console.  However, it does not let you control how \R is
restarted, e.g. with what command-line options and what environment
variables.  Importantly, the new \R session will have the same set of
packages loaded as before, the same variables in the global environment,
and so on.
}

\examples{
\dontrun{
  ## Relaunch R with debugging of startup::startup() enabled
  startup::restart(envvars = c(R_STARTUP_DEBUG = TRUE))

  ## Relaunch R without loading user Rprofile files
  startup::restart(args = "--no-init-file")

  ## Mimic 'R CMD build' and 'R CMD check'
  startup::restart(as = "R CMD build")
  startup::restart(as = "R CMD check")
  ## ... which are both short for
  startup::restart(args = c("--no-restore"),
                   envvars = c(R_DEFAULT_PACKAGES="", LC_COLLATE="C"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_radian.R
\name{is_radian}
\alias{is_radian}
\title{Check if running R via radian (formerly known as rtichoke and rice)}
\usage{
is_radian()
}
\value{
A logical
}
\description{
Check if running R via radian (formerly known as rtichoke and rice)
}
\references{
\enumerate{
\item radian - A 21 century R console (previously known as rtichoke and rice),
\url{https://github.com/randy3k/radian}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/startup_session_options.R
\name{startup_session_options}
\alias{startup_session_options}
\title{Record R session information as options}
\usage{
startup_session_options(action = c("update", "overwrite", "erase"))
}
\arguments{
\item{action}{If \code{"update"} or \code{"overwrite"}, \R options
\code{"startup.session.*"} are set.  If \code{"update"}, then such options that are
not already set are updated.  If \code{"erase"}, any existing
\code{"startup.session.*"} options are removed.}
}
\value{
Returns invisibly a named list of the options prefixed
\code{"startup.session."}:
\describe{
\item{\code{startup.session.startdir}}{(character) the working directory when
the \pkg{startup} was first loaded.  If \code{startup::startup()} is called
at the very beginning of the \file{.Rprofile} file, this is also the
directory that the current \R session was launched from.}
\item{\code{startup.session.starttime}}{(POSIXct) the time when the
\pkg{startup} was first loaded.}
\item{\code{startup.session.id}}{(character) a unique ID for the current \R
session.}
\item{\code{startup.session.dumpto}}{(character) a session-specific name that
can be used for argument \code{dumpto} of \link[utils:debugger]{dump.frames()}
(also for dumping to file).}
}
}
\description{
Record R session information as options
}
\examples{
opts <- startup::startup_session_options()
opts

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/on_session_enter.R, R/on_session_exit.R
\name{on_session_enter}
\alias{on_session_enter}
\alias{on_session_exit}
\title{Register functions to be evaluated at the beginning or end of the R session}
\usage{
on_session_enter(fcn = NULL, append = TRUE, replace = FALSE)

on_session_exit(fcn = NULL, append = TRUE, replace = FALSE)
}
\arguments{
\item{fcn}{A function or an R expression. The function must accept zero
or more arguments (currently not used). If an expression, it will
automatically we wrapped up in an anonymous function.}

\item{append}{If TRUE (default), the function will be evaluated after
previously registered ones, otherwise prepended.}

\item{replace}{if TRUE, the function replaces any previously registered
ones, otherwise it will be added (default).}
}
\value{
(invisible) the list of registered functions.
}
\description{
Register functions to be evaluated at the beginning or end of the R session
}
\details{
These functions register one or more functions to be called when the
current R session begins or ends.  The functions are evaluated in a local
environment and without exception handlers, which means that if one
produces an error, then none of the succeeding functions will be called.

To list currently registered functions, use \code{fcns <- on_session_enter()}
or \code{fcns <- on_session_exit()}.
To remove all registered functions, use \code{on_session_enter(replace = TRUE)}
or \code{on_session_exit(replace = TRUE)}.

The \code{on_session_enter()} function works by recording all \code{fcn}:s in an
internal list which will be evaluated via a custom
\code{\link[base:.First]{.First()}} function created in the global
environment. Any other \code{.First()} function on the search path, including
a pre-existing \code{.First()} function in the global environment, is called
at the end after registered functions have been called.

The \code{on_session_exit()} function works by recording all \code{fcn}:s in an
internal list which will be evaluated via a custom function that is called
when the global environment is garbage collected, which happens at the very
end of the R shutdown process.
Contrary to a \code{\link[base:.Last]{.Last()}} function, which is not be
called if \code{quit(runLast = FALSE)} is used, functions registered via
\code{on_session_exit()} are always processed.
Registered \code{on_session_exit()} functions are called \emph{after} \code{quit()} saves
any workspace image to file (\code{./.RData}), and \emph{after} any \code{.Last()} has
been called.
}
\examples{
\dontrun{
## Summarize interactive session upon termination
if (interactive()) {
  startup::on_session_exit(local({
    t0 <- Sys.time()
    function(...) {
      dt <- difftime(Sys.time(), t0, units = "auto")
      msg <- c(
        "Session summary:",
        sprintf(" * R version: \%s", getRversion()),
        sprintf(" * Process ID: \%d", Sys.getpid()),
        sprintf(" * Wall time: \%.2f \%s", dt, attr(dt, "units"))
      )
      msg <- paste(msg, collapse = "\n")
      message(msg)
    }
  }))
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_files.R
\name{find_rprofile}
\alias{find_rprofile}
\alias{find_renviron}
\title{Locates the .Rprofile and .Renviron files used during the startup of R}
\usage{
find_rprofile(all = FALSE)

find_renviron(all = FALSE)
}
\description{
Locates the .Rprofile and .Renviron files used during the startup of R
}
\section{Functions}{
\itemize{
\item \code{find_rprofile()}: Locates the \file{.Rprofile} file used during
\R startup.

\item \code{find_renviron()}: Locates the \file{.Renviron} file used during
\R startup.

}}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conditions.R
\name{warn}
\alias{warn}
\title{Produce a warning}
\usage{
warn(..., call. = FALSE, immediate. = TRUE, domain = NULL)
}
\arguments{
\item{..., domain}{The message and optionally the domain used for
translation.  The \ldots arguments are passed to \link[base:sprintf]{base::sprintf} to
create the message string.}

\item{call.}{If \link[base:logical]{base::TRUE}, the call is included in the warning
message, otherwise not.}

\item{immediate.}{If \link[base:logical]{base::TRUE}, the warning is outputted immediately,
otherwise not.}
}
\description{
Produce a warning
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_wine.R
\name{is_wine}
\alias{is_wine}
\title{Check if running R on Windows using Wine}
\usage{
is_wine()
}
\value{
A logical
}
\description{
Check if running R on Windows using Wine
}
\references{
\enumerate{
\item Wine Developer FAQ, How can I detect Wine?,
\url{https://wiki.winehq.org/Developer_FAQ#How_can_I_detect_Wine.3F}
\item Jeff Zaroyko, Detecting Wine, Wine Devel mailing list, 2008-09-29,
\url{https://www.winehq.org/pipermail/wine-devel/2008-September/069387.html}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renviron_d.R, R/rprofile_d.R, R/startup.R
\name{renviron_d}
\alias{renviron_d}
\alias{renviron}
\alias{rprofile_d}
\alias{rprofile}
\alias{startup}
\title{Load .Renviron.d and .Rprofile.d directories during the R startup process}
\usage{
renviron_d(
  sibling = FALSE,
  all = FALSE,
  unload = FALSE,
  skip = NA,
  dryrun = NA,
  debug = NA,
  paths = NULL
)

rprofile_d(
  sibling = FALSE,
  all = FALSE,
  check = NA,
  unload = FALSE,
  skip = NA,
  on_error = c("error", "warning", "immediate.warning", "message", "ignore"),
  dryrun = NA,
  debug = NA,
  paths = NULL
)

startup(
  sibling = FALSE,
  all = FALSE,
  on_error = c("error", "warning", "immediate.warning", "message", "ignore"),
  keep = c("options"),
  check = NA,
  unload = TRUE,
  skip = NA,
  encoding = getOption("encoding"),
  dryrun = NA,
  debug = dryrun
)
}
\arguments{
\item{sibling}{If \code{TRUE}, then only \file{.Renviron.d/} and
\file{.Rprofile.d/} directories with a sibling \file{.Renviron} and
\file{.Rprofile} in the same location will be considered.}

\item{all}{If \code{TRUE}, then \emph{all} \file{.Renviron.d/} and \file{.Rprofile.d/}
directories found on \link[base:Startup]{the R startup search path} are
processed, otherwise only the \emph{first ones} found.}

\item{unload}{If \code{TRUE}, then the package is unloaded afterward, otherwise
not.}

\item{skip}{If \code{TRUE}, startup directories will be skipped.  If \code{NA}, they
will be skipped if command-line options \code{--vanilla}, \code{--no-init-file},
and / or \code{--no-environ} were specified.}

\item{dryrun}{If \code{TRUE}, everything is done except the processing of the
startup files.}

\item{debug}{If \code{TRUE}, debug messages are outputted, otherwise not.}

\item{paths}{(internal) character vector of directories.}

\item{check}{If \code{TRUE}, then the content of startup files are validated.}

\item{on_error}{Action taken when an error is detected when sourcing an
Rprofile file.  It is not possible to detect error in Renviron files;
they are always ignored with a message that cannot be captured.}

\item{keep}{Specify what information should remain after this function
complete.  The default is to keep \verb{startup.session.*} options
as recorded by \code{\link[=startup_session_options]{startup_session_options()}}.}

\item{encoding}{The encodingto use when parsing the R startup files.
See \code{\link[base:parse]{base::parse()}} for more details.}
}
\description{
Initiates \R using all files under \file{.Renviron.d/} and / or
\file{.Rprofile.d/} directories (or in subdirectories thereof).
}
\details{
The above is done in addition the \file{.Renviron} and \file{.Rprofile}
files that are supported by the built-in \link[base:Startup]{startup process}
of \R.
}
\section{Functions}{
\itemize{
\item \code{renviron_d()}: Initiate using \file{.Renviron.d/} files

\item \code{rprofile_d()}: Initiate using \file{.Rprofile.d/} files

\item \code{startup()}: \code{renviron_d()} followed by \code{rprofile_d()} and then the
package is unloaded

}}
\section{User-specific installation}{

In order for \file{.Rprofile.d} and \file{.Renviron.d} directories to be
included during the \R startup process, a user needs to add
\code{startup::startup()} to \file{~/.Rprofile}.  Adding this can also be done
by calling \code{\link[=install]{install()}} once.
}

\section{Site-wide installation}{

An alternative to having each user add \code{startup::startup()} in their own
\file{~/.Rprofile} file, is to add it to the site-wide \file{Rprofile.site}
file (see \link[base:Startup]{?Startup}).
The advantage of such a site-wide installation, is that the users do not
have to have a \file{.Rprofile} file for \file{.Rprofile.d} and
\file{.Renviron.d} directories to work.
For this to work for all users automatically, the \pkg{startup} package
should also be installed in the site-wide library.
}

\examples{
\dontrun{
# The most common way to use the package is to add
# the following call to the ~/.Rprofile file.
startup::startup()

# To process ~/.Renviron.d/ files, and then any ./.Renviron.d/ files,
# followed by  ~/.Rprofile.d/ files, and then any ./.Rprofile.d/ files,
# add the following call to the ~/.Rprofile file.
startup::startup(all = TRUE)

# For finer control of on exactly what files are used
# functions renviron_d() and rprofile_d() are also available:

# Initiate first .Renviron.d/ found on search path
startup::renviron_d()

# Initiate all .Rprofile.d/ directories found on the startup search path
startup::rprofile_d(all = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_ess.R
\name{is_ess}
\alias{is_ess}
\title{Check if running R via Emacs Speaks Statistics (ESS)}
\usage{
is_ess()
}
\value{
A logical
}
\description{
Check if running R via Emacs Speaks Statistics (ESS)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install.R
\name{install}
\alias{install}
\alias{uninstall}
\title{Install and uninstall support for .Renviron.d and .Rprofile.d startup
directories}
\usage{
install(
  file = rprofile_user(),
  backup = TRUE,
  overwrite = FALSE,
  path = dirname(file),
  make_dirs = TRUE,
  quiet = FALSE
)

uninstall(file = rprofile_user(), backup = TRUE, quiet = FALSE)
}
\arguments{
\item{file}{The pathname where to create or update the \file{.Rprofile}
file.}

\item{backup}{If \code{TRUE}, a timestamped backup copy of the original file is
created before modifying / overwriting it, otherwise not.  If the backup
fails, then an error is produced and the R startup file is unmodified.}

\item{overwrite}{If the R startup file already exist, then \code{FALSE} (default)
appends the startup code to the end of the file. is overwritten.  If \code{TRUE},
any pre-existing R startup file is overwritten.}

\item{path}{The folder where to create \file{.Renviron.d} and
\file{.Rprofile.d} directory.}

\item{make_dirs}{If \code{TRUE} (default), directories \file{.Renviron.d/} and
\file{.Rprofile.d/} are created in folder \code{path}, if missing.}

\item{quiet}{If \code{FALSE} (default), detailed messages are generated,
otherwise not.}
}
\value{
The pathname of the R startup file modified.
}
\description{
Install and uninstall support for \file{.Renviron.d} and \file{.Rprofile.d}
startup directories by appending / removing one line of code to the
\file{~/.Rprofile} file.
}
\section{Functions}{
\itemize{
\item \code{install()}: injects a \code{tryCatch(startup::startup(), ...)} call to
the \file{.Rprofile} file, which is created if missing.

\item \code{uninstall()}: Remove calls to \code{startup::startup()} and similar.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_rstudio.R
\name{is_rstudio_console}
\alias{is_rstudio_console}
\title{Checks if running R in RStudio Console or RStudio Terminal}
\usage{
is_rstudio_console()
}
\value{
A logical
}
\description{
Checks if running R in RStudio Console or RStudio Terminal
}
\references{
\enumerate{
\item Kevin Ushey (RStudio), Check if R is running in RStudio,
Stackoverflow, 2016-09-23,
\url{https://stackoverflow.com/questions/12389158/check-if-r-is-running-in-rstudio#comment66636507_35849779}
\item Jonathan McPherson (RStudio), Programmatically detect RStudio Terminal vs RStudio Console?,
RStudio Community - RStudio IDE, 2018-01-10,
\url{https://forum.posit.co/t/programmatically-detect-rstudio-terminal-vs-rstudio-console/4107}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{startup.options}
\alias{startup.options}
\alias{R_STARTUP_CHECK}
\alias{startup.check}
\alias{R_STARTUP_CHECK_OPTIONS_IGNORE}
\alias{startup.check.options.ignore}
\alias{startup.commandArgs}
\alias{R_STARTUP_DEBUG}
\alias{startup.debug}
\alias{R_STARTUP_DRYRUN}
\alias{startup.dryrun}
\alias{R_STARTUP_DISABLE}
\alias{startup.disable}
\alias{R_STARTUP_FILE}
\alias{startup.file}
\alias{R_STARTUP_INIT}
\alias{startup.init}
\alias{R_STARTUP_RDATA}
\alias{startup.rdata}
\alias{R_STARTUP_TIME}
\alias{startup.time}
\title{Options and environment variables used by the 'startup' package}
\description{
Below are environment variables and \R options that are used by the
\pkg{startup} package.
The \env{R_STARTUP_***} environment variables must be set before calling
the \code{startup::startup()} function, that is, either (i) prior to launching
\R or (ii) in the \file{.Renviron} file.
}
\section{Controls whether \pkg{startup} is used or not}{


\describe{
\item{\env{R_STARTUP_DISABLE} / \option{startup.disable}:}{
(logical)
If \code{TRUE}, \code{startup::startup()} is fully disable such that \emph{no}
\file{.Renviron.d/} or \file{.Rprofile.d/} files are processed.
\emph{Note}: Files \file{.Renviron} and \file{.Rprofile} are still processed
because these are out of control of the \pkg{startup} package.
(Default: \code{FALSE})
}

\item{\env{R_STARTUP_DRYRUN} / \option{startup.dryrun}:}{
(logical)
Controls the default value of argument \code{dryrun} of \code{\link[=startup]{startup()}}.
(Default: \code{FALSE})
}
}
}

\section{Additional customization of the startup process}{


\describe{
\item{\env{R_STARTUP_FILE} / \option{startup.file}:}{
(R script as a character string)
Optional \R script that is parsed and evaluated after
\file{.Renviron.d/} and \file{.Rprofile.d/} files,
and \code{R_STARTUP_INIT} code, have been processed, e.g.
\verb{R_STARTUP_FILE="setup.R" R --quiet}.
(Default: not specified)
}

\item{\env{R_STARTUP_INIT} / \option{startup.init}:}{
(R code as a character string)
Optional \R code that is parsed and evaluated after
\file{.Renviron.d/} and \file{.Rprofile.d/} files,
but before \code{R_STARTUP_FILE} code, have been processed e.g.
\verb{R_STARTUP_INIT="message('Hello world')" R --quiet}.
The specified string must be parsable by \code{\link[base:parse]{base::parse()}}.
(Default: not specified)
}

\item{\env{R_STARTUP_RDATA} / \option{startup.rdata}:}{
(comma-separated values)
Controls whether an existing \file{./.RData} file should be processed
or not.
If \code{"remove"}, it will be skipped by automatically removing it.
If \code{"rename"}, it will be renamed to \file{./.RData.YYYYMMDD_hhmmss}
where the timestamp is the last time the file was modified.
If \code{"prompt"}, the user is prompted whether they want to load the file
or rename it. In non-interactive session, \code{"prompt"} will fallback to
loading the content (default). To fallback to renaming the file, use
\code{"prompt,rename"}.
Note that in contrast to \code{R} and \verb{R CMD BATCH file.R}, \code{Rscript} does
\emph{not} load \file{.RData} files unless command-line option \code{--restore}
is specified.
(Default: not specified)
}
}
}

\section{Controls what validation checks are performed at startup}{


\describe{
\item{\env{R_STARTUP_CHECK} / \option{startup.check}:}{
(logical)
Controls the default value of argument \code{check} of \code{\link[=startup]{startup()}}.
(Default: \code{TRUE})
}

\item{\env{R_STARTUP_CHECK_OPTIONS_IGNORE} /
\option{startup.check.options.ignore}:}{
(character vector or comma-separated character string)
Names of \R options that should \emph{not} be validated at the end of the
\code{\link[=startup]{startup()}} process.
(Default: \code{"error"})
}
}
}

\section{Settings useful for debugging and prototyping}{


\describe{

\item{\env{R_STARTUP_DEBUG} / \option{startup.debug}:}{
(logical)
Controls the default value of argument \code{debug} of \code{\link[=startup]{startup()}}.
(Default: \code{FALSE})
}

\item{\option{startup.commandArgs}:}{
(character vector)
Overrides the command-line arguments that \code{\link[=startup]{startup()}} uses, which
can be useful to prototype and test alternative ways that \R might
be launched.
(Default: \code{base::commandArgs()})
}

\item{\env{R_STARTUP_TIME} / \option{startup.time}:}{
(POSIX timestamp; character string)
Overrides the current timestamp, which can be useful to prototype and
test functionalities that depend on the current time, e.g. inclusion
and exclusion of files based on \verb{when=<periodicity>} tags.
The specified string must be parsable by \code{\link[base:as.POSIXlt]{base::as.POSIXct()}}.
(Default: not specified)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sysinfo.R
\name{sysinfo}
\alias{sysinfo}
\title{Information on the current R session}
\usage{
sysinfo()
}
\value{
A named list.
}
\description{
Information on the current R session
}
\examples{
startup::sysinfo()

}
